import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.op.functional import *
import pandas as pd
import numpy as np
from abc import ABCMeta, abstractmethod
from torchqtm.tdbt.backtest import BackTestEnv
from torchqtm.base import BaseAlpha, Reversion


class IBS(Reversion):
    def __init__(self, env):
        super().__init__(env)
        self.lag = op.Parameter(5, requires_optim=False, feasible_region=None)

    def forward(self):
        self.data = (self.close - F.ts_min(self.low, self.lag)) / (
                    F.ts_max(self.high, self.lag) - F.ts_min(self.low, self.lag))
        return self.data



import pandas as pd
import numpy as np
from torchqtm.op.functional import *
import time
X_n = np.random.normal(0, 1, (252, 50))
X_n[0, [1, 3, 5]] = np.nan
X_n[2, [1, 3, 5]] = np.nan
X_n[4, [1, 3, 5]] = np.nan
idx = pd.date_range('20100101', periods=252)
symbols = [str(i) for i in range(1, 51)]
X_d = pd.DataFrame(X_n, index=idx, columns=symbols)


class Timer(object):
    def __init__(self, project_name=None):
        self.project_name = project_name
        self.start = time.time()

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("{}, Time elapsed: {:.4f}ms".format(self.project_name, 1000*(time.time() - self.start)))


class Tester(object):
    @staticmethod
    def t_ceiling():
        print(ceiling(X_d))

    @staticmethod
    def t_exp():
        print(np.exp(X_d))

    @staticmethod
    def t_nan_out():
        print(nan_out(X_d))


class CSTester(object):
    @ staticmethod
    def t_cs_corr_0():
        with Timer():
            print(cs_corr(X_d, X_d**2, method='spearman'))

    @ staticmethod
    def t_cs_corr_1():
        def corr(X, Y):
            mean_x = np.nanmean(X, axis=1, keepdims=True)
            std_x = np.nanstd(X, axis=1, keepdims=True)
            mean_y = np.nanmean(Y, axis=1, keepdims=True)
            std_y = np.nanstd(Y, axis=1, keepdims=True)
            return np.nanmean((X-mean_x)/std_x * (Y-mean_y)/std_y, axis=1)
        with Timer():
            print(corr(X_d, X_d**2))

        with Timer():
            print(cs_corr(X_n, X_n**3, method='spearman'))


import numpy as np
import pandas as pd
import time







"""Utilities that are required by gplearn.

Most of these functions are slightly modified versions of some key utility
functions from scikit-learn that gplearn depends upon. They reside here in
order to maintain compatibility across different versions of scikit-learn.

"""

import numbers

import numpy as np
from joblib import cpu_count


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)


def _get_n_jobs(n_jobs):
    """Get number of jobs for the computation.

    This function reimplements the logic of joblib to determine the actual
    number of jobs depending on the cpu count. If -1 all CPUs are used.
    If 1 is given, no parallel computing code is used at all, which is useful
    for debugging. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.
    Thus for n_jobs = -2, all CPUs but one are used.

    Parameters
    ----------
    n_jobs : int
        Number of jobs stated in joblib convention.

    Returns
    -------
    n_jobs : int
        The actual number of jobs as positive integer.

    """
    if n_jobs < 0:
        return max(cpu_count() + 1 + n_jobs, 1)
    elif n_jobs == 0:
        raise ValueError('Parameter n_jobs == 0 has no meaning.')
    else:
        return n_jobs


def _partition_estimators(n_estimators, n_jobs):
    """Private function used to partition estimators between jobs."""
    # Compute the number of jobs
    n_jobs = min(_get_n_jobs(n_jobs), n_estimators)

    # Partition estimators between jobs
    n_estimators_per_job = (n_estimators // n_jobs) * np.ones(n_jobs,
                                                              dtype=int)
    n_estimators_per_job[:n_estimators % n_jobs] += 1
    starts = np.cumsum(n_estimators_per_job)

    return n_jobs, n_estimators_per_job.tolist(), [0] + starts.tolist()



"""Metrics to evaluate the fitness of a program.

The :mod:`gplearn.fitness` module contains some metric with which to evaluate
the computer programs created by the :mod:`gplearn.genetic` module.
"""

# Author: Trevor Stephens <trevorstephens.com>
#
# License: BSD 3 clause

import numbers

import numpy as np
from joblib import wrap_non_picklable_objects
from scipy.stats import rankdata

__all__ = ['make_fitness']


class _Fitness(object):

    """A metric to measure the fitness of a program.

    This object is able to be called with NumPy torchqtm arguments and return
    a resulting floating point score quantifying the quality of the program's
    representation of the true relationship.

    Parameters
    ----------
    function : callable
        A function with signature function(y, y_pred, sample_weight) that
        returns a floating point number. Where `y` is the input target y
        vector, `y_pred` is the predicted values from the genetic program, and
        sample_weight is the sample_weight vector.

    greater_is_better : bool
        Whether a higher value from `function` indicates a better fit. In
        general this would be False for metrics indicating the magnitude of
        the error, and True for metrics indicating the quality of fit.

    """

    def __init__(self, function, greater_is_better):
        self.function = function
        self.greater_is_better = greater_is_better
        self.sign = 1 if greater_is_better else -1

    def __call__(self, *args):
        return self.function(*args)


def make_fitness(*, function, greater_is_better, wrap=True):
    """Make a fitness measure, a metric scoring the quality of a program's fit.

    This factory function creates a fitness measure object which measures the
    quality of a program's fit and thus its likelihood to undergo genetic
    operations into the next generation. The resulting object is able to be
    called with NumPy torchqtm arguments and return a resulting floating
    point score quantifying the quality of the program's representation of the
    true relationship.

    Parameters
    ----------
    function : callable
        A function with signature function(y, y_pred, sample_weight) that
        returns a floating point number. Where `y` is the input target y
        vector, `y_pred` is the predicted values from the genetic program, and
        sample_weight is the sample_weight vector.

    greater_is_better : bool
        Whether a higher value from `function` indicates a better fit. In
        general this would be False for metrics indicating the magnitude of
        the error, and True for metrics indicating the quality of fit.

    wrap : bool, optional (default=True)
        When running in parallel, pickling of custom metrics is not supported
        by Python's default pickler. This option will wrap the function using
        cloudpickle allowing you to pickle your solution, but the evolution may
        run_ slightly more slowly. If you are running single-threaded in an
        interactive Python session or have no need to save the model, set to
        `False` for faster runs.

    """
    # if not isinstance(greater_is_better, bool):
    #     raise ValueError('greater_is_better must be bool, got %s'
    #                      % type(greater_is_better))
    # if not isinstance(wrap, bool):
    #     raise ValueError('wrap must be an bool, got %s' % type(wrap))
    # if function.__code__.co_argcount != 3:
    #     raise ValueError('function requires 3 arguments (y, y_pred, w),'
    #                      ' got %d.' % function.__code__.co_argcount)
    # if not isinstance(function(np.array([1, 1]),
    #                   np.array([2, 2]),
    #                   np.array([1, 1])), numbers.Number):
    #     raise ValueError('function must return a numeric.')
    #
    # if wrap:
    #     return _Fitness(function=wrap_non_picklable_objects(function),
    #                     greater_is_better=greater_is_better)
    return _Fitness(function=function,
                    greater_is_better=greater_is_better)


def _weighted_pearson(y, y_pred, w):
    """Calculate the weighted Pearson correlation coefficient."""
    with np.errstate(divide='ignore', invalid='ignore'):
        y_pred_demean = y_pred - np.average(y_pred, weights=w)
        y_demean = y - np.average(y, weights=w)
        corr = ((np.sum(w * y_pred_demean * y_demean) / np.sum(w)) /
                np.sqrt((np.sum(w * y_pred_demean ** 2) *
                         np.sum(w * y_demean ** 2)) /
                        (np.sum(w) ** 2)))
    if np.isfinite(corr):
        return np.abs(corr)
    return 0.


def _weighted_spearman(y, y_pred, w):
    """Calculate the weighted Spearman correlation coefficient."""
    y_pred_ranked = np.apply_along_axis(rankdata, 0, y_pred)
    y_ranked = np.apply_along_axis(rankdata, 0, y)
    return _weighted_pearson(y_pred_ranked, y_ranked, w)


def _mean_absolute_error(y, y_pred, w):
    """Calculate the mean absolute error."""
    return np.average(np.abs(y_pred - y), weights=w)


def _mean_square_error(y, y_pred, w):
    """Calculate the mean square error."""
    return np.average(((y_pred - y) ** 2), weights=w)


def _root_mean_square_error(y, y_pred, w):
    """Calculate the root mean square error."""
    return np.sqrt(np.average(((y_pred - y) ** 2), weights=w))


def _log_loss(y, y_pred, w):
    """Calculate the log loss."""
    eps = 1e-15
    inv_y_pred = np.clip(1 - y_pred, eps, 1 - eps)
    y_pred = np.clip(y_pred, eps, 1 - eps)
    score = y * np.log(y_pred) + (1 - y) * np.log(inv_y_pred)
    return np.average(-score, weights=w)


weighted_pearson = _Fitness(function=_weighted_pearson,
                            greater_is_better=True)
weighted_spearman = _Fitness(function=_weighted_spearman,
                             greater_is_better=True)
mean_absolute_error = _Fitness(function=_mean_absolute_error,
                               greater_is_better=False)
mean_square_error = _Fitness(function=_mean_square_error,
                             greater_is_better=False)
root_mean_square_error = _Fitness(function=_root_mean_square_error,
                                  greater_is_better=False)
log_loss = _Fitness(function=_log_loss,
                    greater_is_better=False)

_fitness_map = {'pearson': weighted_pearson,
                'spearman': weighted_spearman,
                'mean absolute error': mean_absolute_error,
                'mse': mean_square_error,
                'rmse': root_mean_square_error,
                'log loss': log_loss}



from torchqtm.alphas.autoalpha.gplearn import fitness
from torchqtm.alphas.autoalpha.gplearn import SymbolicRegressor
import numpy as np


def score_func_basic(y, y_pred, sample_weight):  # 适应度函数：策略评价指标
    rlt = np.sum((y - y_pred) ** 2)
    return rlt


if __name__ == '__main__':
    X = np.random.normal(0, 1, size=(252, 50, 6))
    y_ture = X[:, :, -1]

    m = fitness.make_fitness(function=score_func_basic,
                             greater_is_better=False,
                             wrap=False)
    symbolic_model = SymbolicRegressor(population_size=200,
                                       generations=6,
                                       tournament_size=20,
                                       metric=m,
                                       function_set=('add', 'sub'),
                                       feature_names=None,
                                       const_range=(-5000, 5000),
                                       parsimony_coefficient='auto',
                                       stopping_criteria=100,
                                       init_depth=(2, 6),
                                       init_method='half and half',
                                       p_crossover=0.4,
                                       p_subtree_mutation=0.01,
                                       p_hoist_mutation=0.0,
                                       p_point_mutation=0.01,
                                       p_point_replace=0.4,
                                       max_samples=1,
                                       n_jobs=1,
                                       verbose=1,
                                       warm_start=False,
                                       low_memory=False,
                                       random_state=0)

    symbolic_model.fit(X, y_ture)


import pandas as pd
import numpy as np
from torch.utils.data import Dataset, DataLoader
DataLoader()
# This section, we develop a data loader for tseries learning

# Suppose we have features X and labels y with same length

# we use 60% for training, 20% for validation, 20% for testing

# and then, move the training window to the next.

N = 10000
batch_number = 7

# 我们需要移动6次, 也就是 batch_size + 6 * 0.2 * batch_size = N

# suppose Now, we have the dataset, that is, we can access the data using __getitem__

# now we create a data loader


class TSDataLoader0(object):
    def __init__(self, dataset, batch_size=7, step_ratio=0.2):
        self.dataset = dataset
        self.batch_size = batch_size
        self.step_ratio = step_ratio

        self.len_batch = len(dataset) // (1 + step_ratio)
        self.current_index = self.len_batch
        self.step_size = self.len_batch * step_ratio

    # def _get_iterator(self):
    #     pass

    def __iter__(self):
        return self

    def __next__(self):
        train_batch = self.dataset[: self.current_index - 2 * self.step_size]
        validation_batch = self.dataset[self.current_index - 2 * self.step_size: self.current_index - self.step_size]
        test_batch = self.dataset[self.current_index - self.step_size: self.current_index]

        self.current_index += self.step_size
        if self.current_index > len(self.dataset):
            raise StopIteration

        return train_batch, validation_batch, test_batch


class TSDataLoader1(object):
    def __init__(self, dataset, batch_size=7, step_ratio=0.2):
        self.dataset = dataset
        self.batch_size = batch_size
        self.step_ratio = step_ratio

        self.len_batch = len(dataset) // (1 + step_ratio)
        self.current_index = self.len_batch
        self.step_size = self.len_batch * step_ratio

    # def _get_iterator(self):
    #     pass

    def __iter__(self):
        return self

    def __next__(self):
        train_batch = self.dataset[self.current_index-self.len_batch: self.current_index - 2 * self.step_size]
        validation_batch = self.dataset[self.current_index - 2 * self.step_size: self.current_index - self.step_size]
        test_batch = self.dataset[self.current_index - self.step_size: self.current_index]

        self.current_index += self.step_size
        if self.current_index > len(self.dataset):
            raise StopIteration

        return train_batch, validation_batch, test_batch





from typing import Callable
from torchqtm.types import F
from functools import wraps


def ContextManager(manager, used=True) -> Callable[[F], F]:
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if used:
                with manager:
                    func(*args, **kwargs)
            else:
                func(*args, **kwargs)
        return wrapper
    return decorator










import pandas as pd
import tushare as ts
import numpy as np
from torchqtm.config import __TS_API__
from torchqtm.utils import relativedelta, datetime
pro = ts.pro_api(__TS_API__)
from typing import Iterable, Union, List
from abc import abstractmethod, ABCMeta
import joblib


class Universe(object, metaclass=ABCMeta):
    def __init__(self):
        pass

    @abstractmethod
    def get_symbols(self, trade_date: Union[str, datetime, pd.Timestamp]):
        raise NotImplementedError

    @abstractmethod
    def data(self):
        pass


class StaticUniverse(Universe):
    def __init__(self, symbols: Iterable[str]):
        super().__init__()
        self.symbols = symbols

    def get_symbols(self, trade_date=None):
        return self.symbols

    @property
    def data(self):
        return self.symbols


class DynamicUniverse(object):
    pass


# @joblib.Memory('./.cache', verbose=0).cache
class IndexComponents(object):
    def __init__(self,
                 index_code: str,
                 in_date: str):
        self.index_code = index_code
        self.in_date = in_date
        self.data = self._get_components()

    def _get_components(self) -> List[str]:
        end = datetime.strptime(self.in_date, "%Y%m%d")
        start = end - relativedelta(months=1)
        df = pro.index_weight(index_code=self.index_code, start_date=start.strftime("%Y%m%d"), end_date=end.strftime("%Y%m%d"))
        rlt = df[df['trade_date'] == max(np.unique(df['trade_date']))]['con_code']
        rlt = sorted(list(rlt.apply(lambda x: x[:-3])))
        return rlt



from torchqtm.assets import Asset
from torchqtm.finance.transaction import Transaction
import pandas as pd
import math

import logging

log = logging.getLogger("Performance")


class Position:
    def __init__(
            self,
            asset: Asset,
            amount: int = 0,
            cost_basis: float = .0,
            last_sale_price: float = .0,
            last_sale_date: pd.Timestamp = None
    ):
        self.asset = asset
        self.amount = amount
        self.cost_basis = cost_basis
        self.last_sale_price = last_sale_price
        self.last_sale_date = last_sale_date

    def earn_dividend(self, dividend):
        """
        Register the number of shares we held at this dividend's ex date so
        that we can pay out the correct amount on the dividend's pay date.
        """
        return {"amount": self.amount * dividend.amount}

    # TODO: implement this
    def earn_stock_dividend(self):
        pass

    def handle_split(self, asset: Asset, ratio: float) -> float:
        """
        在分股时会产生小数, 未了在会计上保持一致, 我们return 这部分现金
        """
        assert self.asset == asset

        raw_share = self.amount / ratio

        true_share_count = math.floor(raw_share)

        fractional_share = raw_share - true_share_count

        new_cost_basis = round(self.cost_basis * ratio, 2)

        self.cost_basis = new_cost_basis
        self.amount = true_share_count

        returned_cash = round(float(fractional_share * new_cost_basis), 2)

        log.info("after split: " + str(self))
        log.info("returning cash: " + str(returned_cash))
        return returned_cash

    def handle_transaction(self, txn: Transaction) -> None:
        assert self.asset == txn.asset

        total_amount = self.amount + txn.amount

        if total_amount == 0:
            self.cost_basis = 0.0
        else:
            pre_direction = math.copysign(1, self.amount)
            txn_direction = math.copysign(1, txn.amount)

            if txn_direction != pre_direction:
                if abs(txn.amount) > abs(self.amount):
                    self.cost_basis = txn.price
            else:
                pre_cost = self.cost_basis * self.amount
                txn_cost = txn.price * txn.amount
                total_cost = pre_cost + txn_cost
                self.cost_basis = total_cost / total_amount

            # Update the last sale price if txn is best data we so far
            if self.last_sale_date is None or txn.dt > self.last_sale_date:
                self.last_sale_price = txn.price
                self.last_sale_date = txn.dt

        self.amount = total_amount

    # TODO: implement this
    def handle_commission(self, asset: Asset, cost: float) -> None:
        """

        """
        pass

    def __repr__(self):
        template = """Position(asset: {asset}, amount: {amount}, cost_basis: {cost_basis}, last_sale_price: {last_sale_price})
        """
        return template.format(
            asset=self.asset,
            amount=self.amount,
            cost_basis=self.cost_basis,
            last_sale_price=self.last_sale_price
        )

    def to_dict(self):
        return {
            "sid": self.asset,
            "amount": self.amount,
            "cost_basis": self.cost_basis,
            "last_sale_price": self.last_sale_price
        }







from abc import ABCMeta, abstractmethod

from torchqtm.types import EVENT_TYPE


class CancelPolicy(object, metaclass=ABCMeta):

    @abstractmethod
    def should_cancel(self, event: EVENT_TYPE):
        raise NotImplementedError


class EODCancel(CancelPolicy):
    """
    This policy cancels open get_orders at the end of the day.
    It is highly recommended that only apply this policy to minutely simulations.
    """
    def __init__(self, warn_on_cancel: bool = True):
        self.warn_on_cancel = warn_on_cancel

    def should_cancel(self, event: EVENT_TYPE):
        return event == EVENT_TYPE.SESSION_END


class NeverCancel(CancelPolicy):
    """
    Order are never automatically canceled.
    """
    def __init__(self):
        self.warn_on_cancel = False

    def should_cancel(self, event: EVENT_TYPE):
        return False




import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.op.functional import *
import pandas as pd
import numpy as np
from abc import ABCMeta, abstractmethod
from torchqtm.tdbt.backtest import BackTestEnv
from torchqtm.base import BaseAlpha


class RSI(op.Technical):
    def __init__(self, env):
        super().__init__(env)
        self.lag = op.Parameter(5, requires_optim=False, feasible_region=None)

    def forward(self):
        self.data = self.env.Close
        return self.data


class BollingerBands(op.Technical):
    pass


class Aroon(op.Technical):
    pass


class MACD(op.Technical):
    pass


# class FastStochasticOscillator(CustomFactor):
#     """
#     Fast Stochastic Oscillator Indicator [%K, Momentum Indicator]
#     https://wiki.timetotrade.eu/Stochastic
#
#     This stochastic is considered volatile, and varies a lot when used in
#     market analysis. It is recommended to use the slow stochastic oscillator
#     or a moving average of the %K [%D].
#
#     **Default Inputs:** :data:`zipline.pipeline.data.EquityPricing.close`, \
#                         :data:`zipline.pipeline.data.EquityPricing.low`, \
#                         :data:`zipline.pipeline.data.EquityPricing.high`
#
#     **Default Window Length:** 14
#
#     Returns
#     -------
#     out: %K oscillator
#     """
#     inputs = (EquityPricing.close, EquityPricing.low, EquityPricing.high)
#     window_safe = True
#     window_length = 14
#
#     def compute(self, today, assets, out, closes, lows, highs):
#
#         highest_highs = nanmax(highs, axis=0)
#         lowest_lows = nanmin(lows, axis=0)
#         today_closes = closes[-1]
#
#         evaluate(
#             '((tc - ll) / (hh - ll)) * 100',
#             local_dict={
#                 'tc': today_closes,
#                 'll': lowest_lows,
#                 'hh': highest_highs,
#             },
#             global_dict={},
#             out=out,
#         )
#
#
# class IchimokuKinkoHyo(CustomFactor):
#     """Compute the various metrics for the Ichimoku Kinko Hyo (Ichimoku Cloud).
#     http://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:ichimoku_cloud
#
#     **Default Inputs:** :data:`zipline.pipeline.data.EquityPricing.high`, \
#                         :data:`zipline.pipeline.data.EquityPricing.low`, \
#                         :data:`zipline.pipeline.data.EquityPricing.close`
#
#     **Default Window Length:** 52
#
#     Parameters
#     ----------
#     window_length : int > 0
#         The length the the window for the senkou span b.
#     tenkan_sen_length : int >= 0, <= window_length
#         The length of the window for the tenkan-sen.
#     kijun_sen_length : int >= 0, <= window_length
#         The length of the window for the kijou-sen.
#     chikou_span_length : int >= 0, <= window_length
#         The lag for the chikou span.
#     """ # noqa
#
#     params = {
#         'tenkan_sen_length': 9,
#         'kijun_sen_length': 26,
#         'chikou_span_length': 26,
#     }
#     inputs = (EquityPricing.high, EquityPricing.low, EquityPricing.close)
#     outputs = (
#         'tenkan_sen',
#         'kijun_sen',
#         'senkou_span_a',
#         'senkou_span_b',
#         'chikou_span',
#     )
#     window_length = 52
#
#     def _validate(self):
#         super(IchimokuKinkoHyo, self)._validate()
#         for k, v in self.params.items():
#             if v > self.window_length:
#                 raise ValueError(
#                     '%s must be <= the window_length: %s > %s' % (
#                         k, v, self.window_length,
#                     ),
#                 )
#
#     def compute(self,
#                 today,
#                 assets,
#                 out,
#                 high,
#                 low,
#                 close,
#                 tenkan_sen_length,
#                 kijun_sen_length,
#                 chikou_span_length):
#
#         out.tenkan_sen = tenkan_sen = (
#             high[-tenkan_sen_length:].max(axis=0) +
#             low[-tenkan_sen_length:].min(axis=0)
#         ) / 2
#         out.kijun_sen = kijun_sen = (
#             high[-kijun_sen_length:].max(axis=0) +
#             low[-kijun_sen_length:].min(axis=0)
#         ) / 2
#         out.senkou_span_a = (tenkan_sen + kijun_sen) / 2
#         out.senkou_span_b = (high.max(axis=0) + low.min(axis=0)) / 2
#         out.chikou_span = close[chikou_span_length]
#
#
# class RateOfChangePercentage(CustomFactor):
#     """
#     Rate of change Percentage
#     ROC measures the percentage change in price from one period to the next.
#     The ROC calculation compares the current price with the price `n`
#     periods ago.
#     Formula for calculation: ((price - prevPrice) / prevPrice) * 100
#     price - the current price
#     prevPrice - the price n days ago, equals window length
#     """
#     def compute(self, today, assets, out, close):
#         today_close = close[-1]
#         prev_close = close[0]
#         evaluate('((tc - pc) / pc) * 100',
#                  local_dict={
#                      'tc': today_close,
#                      'pc': prev_close
#                  },
#                  global_dict={},
#                  out=out,
#                  )
#
#
# class TrueRange(CustomFactor):
#     """
#     True Range
#
#     A technical indicator originally developed by J. Welles Wilder, Jr.
#     Indicates the true degree of daily price change in an underlying.
#
#     **Default Inputs:** :data:`zipline.pipeline.data.EquityPricing.high`, \
#                         :data:`zipline.pipeline.data.EquityPricing.low`, \
#                         :data:`zipline.pipeline.data.EquityPricing.close`
#
#     **Default Window Length:** 2
#     """
#     inputs = (
#         EquityPricing.high,
#         EquityPricing.low,
#         EquityPricing.close,
#     )
#     window_length = 2
#
#     def compute(self, today, assets, out, highs, lows, closes):
#         high_to_low = highs[1:] - lows[1:]
#         high_to_prev_close = abs(highs[1:] - closes[:-1])
#         low_to_prev_close = abs(lows[1:] - closes[:-1])
#         out[:] = nanmax(
#             dstack((
#                 high_to_low,
#                 high_to_prev_close,
#                 low_to_prev_close,
#             )),
#             2
#         )
#
#
# class MovingAverageConvergenceDivergenceSignal(CustomFactor):
#     """
#     Moving Average Convergence/Divergence (MACD) Signal line
#     https://en.wikipedia.org/wiki/MACD
#
#     A technical indicator originally developed by Gerald Appel in the late
#     1970's. MACD shows the relationship between two moving averages and
#     reveals changes in the strength, direction, momentum, and duration of a
#     trend in a stock's price.
#
#     **Default Inputs:** :data:`zipline.pipeline.data.EquityPricing.close`
#
#     Parameters
#     ----------
#     fast_period : int > 0, optional
#         The window length for the "fast" EWMA. Default is 12.
#     slow_period : int > 0, > fast_period, optional
#         The window length for the "slow" EWMA. Default is 26.
#     signal_period : int > 0, < fast_period, optional
#         The window length for the signal line. Default is 9.
#
#     Notes
#     -----
#     Unlike most pipeline expressions, this factor does not accept a
#     ``window_length`` parameter. ``window_length`` is inferred from
#     ``slow_period`` and ``signal_period``.
#     """
#     inputs = (EquityPricing.close,)
#     # We don't use the default form of `params` here because we want to
#     # dynamically calculate `window_length` from the period lengths in our
#     # __new__.
#     params = ('fast_period', 'slow_period', 'signal_period')
#
#     @expect_bounded(
#         __funcname='MACDSignal',
#         fast_period=(1, None),  # These must all be >= 1.
#         slow_period=(1, None),
#         signal_period=(1, None),
#     )
#     def __new__(cls,
#                 fast_period=12,
#                 slow_period=26,
#                 signal_period=9,
#                 *args,
#                 **kwargs):
#
#         if slow_period <= fast_period:
#             raise ValueError(
#                 "'slow_period' must be greater than 'fast_period', but got\n"
#                 "slow_period={slow}, fast_period={fast}".format(
#                     slow=slow_period,
#                     fast=fast_period,
#                 )
#             )
#
#         return super(MovingAverageConvergenceDivergenceSignal, cls).__new__(
#             cls,
#             fast_period=fast_period,
#             slow_period=slow_period,
#             signal_period=signal_period,
#             window_length=slow_period + signal_period - 1,
#             *args, **kwargs
#         )
#
#     def _ewma(self, data, length):
#         decay_rate = 1.0 - (2.0 / (1.0 + length))
#         return average(
#             data,
#             axis=1,
#             weights=exponential_weights(length, decay_rate)
#         )
#
#     def compute(self, today, assets, out, close, fast_period, slow_period,
#                 signal_period):
#         slow_EWMA = self._ewma(
#             rolling_window(close, slow_period),
#             slow_period
#         )
#         fast_EWMA = self._ewma(
#             rolling_window(close, fast_period)[-signal_period:],
#             fast_period
#         )
#         macd = fast_EWMA - slow_EWMA
#         out[:] = self._ewma(macd.T, signal_period)









import pandas as pd

import torchqtm.core.window.rolling as F
from torchqtm.utils import Timer
import numpy as np
a = np.random.normal(0, 1, (252 * 10, 5000, 2))
a[..., 1] = -a[..., 0]

pd.DataFrame(a[..., 0]).rolling(10).corr(pd.DataFrame(a[..., 0]))

def ff(x):
    x1 = x[..., 0]
    x2 = x[..., 1]
    rlt = np.sum(x1, axis=0) + np.sum(x2, axis=0)
    return rlt


ff(a).shape
F.roll_apply(a, 5, ff)


m = pd.DataFrame([1,2,3,np.nan,5,6])
b = [0,0,0, 1,1, np.nan]
m.groupby(b).apply(lambda x: np.mean(x))
from scipy.stats import norm

with Timer():
    pd.DataFrame(a[..., 0]).rolling(10).corr(pd.DataFrame(a[..., 0]))


def aux_func(data_slice):
    x = data_slice[..., 0]
    y = data_slice[..., 1]
    retur


with Timer():
    pd.DataFrame(a).rolling(5).rank()

F.roll_apply_rank(a, 10, mode="roll_single")

with Timer():
    F.roll_apply_rank(a, 20, mode="auto")


with Timer("rolling_apply"):
    print(F.roll_apply(a, 500, lambda x: np.max(x, axis=0))[-3:, :5])

with Timer("rolling_apply"):
    print(F.roll_apply_max(a, 5))


with Timer("rolling_apply"):
    print(pd.DataFrame(a).rolling(500).mean().iloc[-3:, :5])

with Timer("rolling_apply"):
    print(pd.DataFrame(a).rolling(5).max().iloc[-3:, :5])


import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.op.functional import *
import pandas as pd
import numpy as np
from abc import ABCMeta, abstractmethod
from torchqtm.tdbt.backtest import BackTestEnv
from torchqtm.base import BaseAlpha


class IntroductoryAlpha(BaseAlpha, metaclass=ABCMeta):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'IntroductoryAlpha'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Step01(IntroductoryAlpha):
    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        adv20 = ts_mean(self.volume, 20)
        event = geq(self.volume, adv20)
        alpha_if_true = 2*(-ts_delta(self.close, 3))
        alpha_if_false = (-ts_delta(self.close, 3))
        self.data = if_else(event, alpha_if_true, alpha_if_false)
        return self.data


class Step02(IntroductoryAlpha):
    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        adv20 = ts_mean(self.volume, 20)
        event = geq(self.volume, adv20)
        alpha = (-ts_delta(self.close, 3))
        self.data = trade_when(event, alpha, False)
        return self.data




import pandas as pd
import numpy as np
from torchqtm.constants import GLOBAL_DEFAULT_START


class DateTimeManager:
    def __init__(self, trading_calendar):
        self.trading_calendar = trading_calendar
        self.datetime: pd.Timestamp = GLOBAL_DEFAULT_START

    def set_dt(self, dt: pd.Timestamp):
        if isinstance(dt, pd.Timestamp):
            assert self.datetime <= dt, "Reverse backtest is not allowed"
            self.datetime = dt
        else:
            raise TypeError("dt should be a pd.Timestamp object")

    @property
    def current_dt(self):
        return self.datetime

    # def advance_time(self, delta):
    #     # delta should be a pd.DateOffset object
    #     if isinstance(delta, pd.DateOffset):
    #         self._current_time += delta
    #     else:
    #         raise TypeError("delta should be a pd.DateOffset object")


class DateTimeMixin:
    def __init__(self, dt_manager):
        self.datetime_manager = dt_manager

    @property
    def current_dt(self):
        return self.datetime_manager.current_dt

    def set_dt(self, dt):
        self.datetime_manager.set_dt(dt)

# TradingAlgorithm, Account, DataPortal


def convert_date_to_int(dt):
    t = dt.year * 10000 + dt.month * 100 + dt.day
    t *= 1e6
    return t


def convert_dates_to_int(dates: pd.DatetimeIndex):
    intDates = dates.year * 10000 + dates.month * 100 + dates.day
    intDates *= 1e6
    return intDates.values.astype(np.uint64)


from datahandler import BarData
from queue import Queue
import pandas as pd
import numpy as np


class VirtualAccount(object):
    def __init__(self, name, bars: BarData):
        self.name = name
        self.bars = bars
        self.auto_update = True
        self.fills = Queue()
        self.weights = []

    def order_pct_to(self, symbols, weights) -> None:
        assert len(symbols) == len(weights), "the length of symbols and weights must be equal"
        assert np.abs(np.sum(weights) - 1.0) <= 0.001, "the sum of weights must be 1"
        assert self.fills.empty(), "一条bar上仅能设置一个仓位"
        target_weights = self.bars.get_template_cseries
        target_weights.loc[symbols] = weights
        target_weights.fillna(0, inplace=True)
        self.fills.put(target_weights)  # 往订单中放入目标权重
        self.auto_update = False

    def _reset_account(self):  # 系统内部每天调用一次
        self.auto_update = True

    def _passive_update(self):
        if not self.auto_update:
            return None
        if len(self.weights) == 0:
            self.fills.put(self.bars.get_template_cseries.fillna(0))
        else:
            weights_rebalanced = self.weights[-1].copy(deep=True)
            weights_rebalanced = weights_rebalanced * (1 + self.bars.history('PctChange'))
            if np.sum(weights_rebalanced) != 1 and np.sum(weights_rebalanced) != 0:
                weights_rebalanced = weights_rebalanced / np.sum(weights_rebalanced)
            self.fills.put(weights_rebalanced)
        self.auto_update = False

    def _update_from_fills(self):
        assert not self.fills.empty(), "the fills queue is empty"
        self.weights.append(self.fills.get())

    def update_account(self):
        self._passive_update()
        self._update_from_fills()
        self._reset_account()

    def calculate_returns(self):
        weights = pd.concat(self.weights, axis=1).T
        weights.index = self.bars.time_index
        weights = weights.shift(1).fillna(0)
        returns = weights * self.bars.PctChange
        returns = returns.sum(axis=1)
        return returns




from copy import copy
import typing
import pandas as pd
from torchqtm.finance.orders.order import Order
from torchqtm.assets import Asset


class Transaction:
    def __init__(
            self,
            asset: Asset,
            amount: int,
            dt: pd.Timestamp,
            price: float,
            order_id: str,
    ):
        self.asset = asset
        self.amount = amount
        self.dt = dt
        self.price = price
        self.order_id = order_id
        # TODO: implement type
        self.type = None

    def __getitem__(self, item):
        return self.__dict__[item]

    def __repr__(self):
        template = "{cls}(asset={asset}, dt={dt}," " amount={amount}, price={price})"

        return template.format(
            cls=type(self).__name__,
            asset=self.asset,
            dt=self.dt,
            amount=self.amount,
            price=self.price,
        )

    def to_dict(self):
        # TODO: implement this
        pass


def create_transaction(order, amount: int, dt: pd.Timestamp, price: float):
    """Create a transaction through a given order
    我感觉没必要用这个函数
    """
    transaction = Transaction(
        asset=order.asset,
        amount=amount,
        dt=dt,
        price=price,
        order_id=order.id
    )
    return transaction



import copy

import numpy as np
import pandas as pd
from collections import deque
from typing import Union
import multiprocessing as mp
import concurrent.futures
from joblib import Parallel, delayed
import os
import sys
import time
import pickle

# TODO: 增加数据源


class BarData(object):
    def __init__(self, datas, start_date, end_date, lookback: int):
        """
        如果没有OHLCV数据, 那么请用0进行填充
        :param datas:
        """
        # assert 'Open' in datas, 'Open is required'
        # assert 'High' in datas, 'High is required'
        # assert 'Low' in datas, 'Low is required'
        # assert 'Close' in datas, 'Close is required'
        # assert 'Volume' in datas, 'Volume is required'
        assert isinstance(lookback, int)
        self.datas = datas
        self.start_date = start_date
        self.end_date = end_date
        if start_date is not None and end_date is not None:
            for key in self.datas:
                self.datas[key] = self.datas[key].loc[start_date:end_date]
        # TODO: 这个pctchange衡量了交易的价格
        self.datas['PctChange'] = self.datas['Close'].pct_change().fillna(0)
        self.PctChange = ...
        self.__create_PctChange()
        self.lookback = lookback
        self.time_index = self.datas['Close'].index
        self.attrs = list(self.datas.keys())
        self.stocks = self.datas['Close'].columns

        self.continue_backtest = True  # 也可以看作, 当天是否取出了数据
        self.bar_index = -1
        self.time_buffers, self.data_buffers = self._create_buffers()
        # self.__save_structure_data()
        self._convert_datas()
        print('success')

    def __create_PctChange(self) -> None:
        self.PctChange = self.datas['Close'].pct_change().fillna(0)

    def __save_structure_data(self):
        with open('modified_datas.pkl', 'wb') as f:
            temp = copy.deepcopy(self.datas)
            for key in self.datas:
                temp[key] = list(self.datas[key].iterrows())
            pickle.dump(temp, f)

    def _create_buffers(self):
        return deque(maxlen=self.lookback), {key: deque(maxlen=self.lookback) for key in self.datas.keys()}

    def _convert_datas(self):
        from queue import Queue
        for key in self.datas:
            self.datas[key] = self.datas[key].iterrows()

    def update_bars(self) -> None:
        """
        Pushes the latest bar to the history_symbol_data structure
        for all symbols in the self.symbol_list
        """

        self.bar_index += 1
        try:
            for key in self.datas:
                current_bar = next(self.datas[key])
                if key == 'Close':
                    self.time_buffers.append(current_bar[0])
                else:
                    self.data_buffers[key].append(current_bar[1])

        except StopIteration:
            self.continue_backtest = False

    def history(self, attr, N=1) -> Union[pd.DataFrame, pd.Series]:
        if N == 1:
            return self.data_buffers[attr][-1]
        else:
            return pd.concat(list(self.data_buffers[attr])[-N:], axis=1).T

    @property
    def now(self):
        return self.time_buffers[-1]

    @property
    def get_template_frame(self):
        return pd.DataFrame(np.nan, index=self.time_index, columns=self.stocks)

    @property
    def get_template_tseries(self):
        return pd.Series(np.nan, index=self.time_index)

    @property
    def get_template_cseries(self):
        return pd.Series(np.nan, index=self.stocks)


def update_bars_process(data, key, time_buffers, data_buffers):
    try:
        current_bar = next(data)
        if key == 'Close':
            time_buffers.put(current_bar[0])
        else:
            data_buffers[key].put(current_bar[1])

    except StopIteration:
        pass


class MultiBarData(BarData):
    def __convert_dataframe(self, key):
        return self.datas[key].iterrows()

    def update_bars(self) -> None:
        """
        Pushes the latest bar to the history_symbol_data structure
        for all symbols in the self.symbol_list
        """

        self.bar_index += 1

        processes = []
        for key in self.datas:
            process = mp.Process(target=update_bars_process, args=(self.datas[key], key, self.time_buffers, self.data_buffers))
            processes.append(process)
            process.start()

        for process in processes:
            process.join()

        try:
            for key in self.data_buffers:
                while not self.data_buffers[key].empty():
                    self.data_buffers_list[key].append(self.data_buffers[key].get())

        except StopIteration:
            self.continue_backtest = False


def update_bars_process(data, key, time_buffers, data_buffers):
    try:
        current_bar = next(data)
        if key == 'Close':
            time_buffers.put(current_bar[0])
        else:
            data_buffers[key].put(current_bar[1])

    except StopIteration:
        pass


class ModifiedBarData(BarData):
    def __init__(self, datas, start_date, end_date, lookback: int):
        """
        如果没有OHLCV数据, 那么请用0进行填充
        :param datas:
        """
        # assert 'Open' in datas, 'Open is required'
        # assert 'High' in datas, 'High is required'
        # assert 'Low' in datas, 'Low is required'
        # assert 'Close' in datas, 'Close is required'
        # assert 'Volume' in datas, 'Volume is required'
        assert isinstance(lookback, int)
        self.datas = datas
        self.start_date = start_date
        self.end_date = end_date
        # TODO: 这个pctchange衡量了交易的价格
        self.PctChange = ...
        self.__create_PctChange()
        self.lookback = lookback
        self.time_index = pd.DatetimeIndex([x[0] for x in self.datas['Close']])
        self.attrs = list(self.datas.keys())

        self.continue_backtest = True  # 也可以看作, 当天是否取出了数据
        self.bar_index = -1
        self.time_buffers, self.data_buffers = self._create_buffers()
        self.stocks = self.datas['Close'][0][1].index
        # self.__save_structure_data()
        self._convert_datas()
        print('success')

    def __create_PctChange(self) -> None:
        self.PctChange = pd.concat([x[1] for x in self.datas['Close']], axis=1).T.pct_change().fillna(0)

    def _convert_datas(self):
        return None

    def update_bars(self) -> None:
        """
        Pushes the latest bar to the history_symbol_data structure
        for all symbols in the self.symbol_list
        """

        self.bar_index += 1
        try:
            for key in self.datas:
                current_bar = self.datas[key].pop(0)
                if key == 'Close':
                    self.time_buffers.append(current_bar[0])
                else:
                    self.data_buffers[key].append(current_bar[1])

        except IndexError:
            self.continue_backtest = False



from abc import ABCMeta, abstractmethod
from torchqtm.utils.universe import Universe
import pandas as pd
import numpy as np
from typing import Dict, Hashable
from abc import ABCMeta, abstractmethod
from typing import Iterable, Callable
from collections import OrderedDict
from torchqtm.utils._decorators import ContextManager
from torchqtm.utils.warnings import catch_warnings


class BackTestEnv(object):
    """
    # We can create two different BackTestEnvs
    # 1. For Op, we create env with trade_dates
    # 2. For backtest, we create env with rebalance_dates
    """

    def __init__(self,
                 dfs: Dict[Hashable, pd.DataFrame],
                 dates,
                 symbols):
        """
        :param dfs:
        :param dates:
        :param symbols: 构造的表的columns, 和Universe的概念不同, 我们的因子可以在更大的范围内计算, 但是只在universe上进行回测
        """
        assert isinstance(dfs, dict)
        self.dfs = dfs
        self._check_dfs()
        self.symbols = symbols
        self.dates = dates
        self.shape = (len(self.symbols), len(self.dates))
        self.Open = None
        self.High = None
        self.Low = None
        self.Close = None
        self.Volume = None
        self.Returns = None
        self.Vwap = None
        self.MktVal = None
        self.PE = None
        self.Sector = None
        self.forward_returns = None
        self._create_datas()
        self._create_features()

    def _check_dfs(self):
        """
        check is dfs is a valid dict
        :return: None
        """
        # assert 'close' in self.dfs
        assert 'MktVal' in self.dfs
        assert 'Sector' in self.dfs

    def _create_datas(self):
        self.data = {}
        for key in self.dfs:
            if isinstance(self.dfs[key], pd.DataFrame):
                self.data[key] = self.dfs[key].loc[self.dates, self.symbols]
            else:
                self.data[key] = self.dfs[key]

    def create_forward_returns(self, D=1):
        forward_returns = self.data['Close'].pct_change().shift(-1)
        return forward_returns

    def _create_features(self):
        """
        Create the reference to the dict values
        :return:
        """
        for key in self.data.keys():
            setattr(self, key, self.data[key])

    def __getitem__(self, item):
        """
        Keep the operator[]
        :param item:
        :return:
        """
        return self.data[item]

    def __setitem__(self, item, value):
        """
        Keep the operator[]
        :param item:
        :return:
        """
        assert isinstance(value, pd.DataFrame)
        self.data[item] = value
        setattr(self, item, value)

    def __delitem__(self, item):
        del self.data[item]
        delattr(self, item)

    def __contains__(self, item):
        return item in self.data

    def match(self, factor):
        return factor.loc[self.dates, self.symbols]


class Parameter(np.ndarray, metaclass=ABCMeta):
    def __new__(cls,
                data: int,
                requires_optim: bool = False,
                feasible_region: Iterable[int] = None):
        obj = np.asarray(data).view(cls)
        obj.required_optim = requires_optim
        obj.feasible_region = feasible_region
        obj._check_data()
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.required_optim = getattr(obj, 'required_optim', None)
        self.feasible_region = getattr(obj, 'feasible_region', None)

    def _check_data(self):
        if self.required_optim is True and self.feasible_region is None:
            raise ValueError("must provide possible values for search")


__TYPES = ['momentum', 'reversion']


class BaseOperator(object, metaclass=ABCMeta):

    catch_warnings = False

    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)


class BaseAlpha(BaseOperator, metaclass=ABCMeta):

    __DATA_AVAILABLE__ = ['open', 'high', 'low', 'close', 'volume',
                          'returns', 'vwap', 'adv20', 'sector']

    def __init__(self, env: BackTestEnv, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.env = env
        self.args = args
        self.kwargs = kwargs
        self.data = None
        self.type = None
        self.Enabled = True  # default to be used

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError

    def __eq__(self, other):
        return self.__repr__() == other.__repr__() and self.data == other.data

    @property
    def open(self):
        return self.env.Open

    @property
    def high(self):
        return self.env.High

    @property
    def low(self):
        return self.env.Low

    @property
    def close(self):
        return self.env.Close

    @property
    def volume(self):
        return self.env.Volume

    @property
    def returns(self):
        return self.env.Returns

    @property
    def vwap(self):
        return self.env.Vwap

    @property
    def adv20(self):
        return self.env.adv20

    @property
    def sector(self):
        return self.env.Sector


class Volatility(BaseAlpha):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'volatility'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Fundamental(BaseAlpha):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'fundamental'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Momentum(BaseAlpha):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'momentum'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Reversion(BaseAlpha):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'reversion'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Technical(BaseAlpha):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'technical'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError







# TODO: implement it

from collections import OrderedDict
from functools import partial
import math
import copy
import logging
import numpy as np
import pandas as pd
from torchqtm.assets import Asset
from torchqtm.finance.transaction import Transaction
from torchqtm.finance.positions.position import Position
# from zipline.utils.sentinel import sentinel
import typing
from torchqtm.data.data_portal import DataPortal
from torchqtm.assets import Future
# from torchqtm.finance._finance_ext import PositionStats
# calculate_positions_tracker_stats
log = logging.getLogger("PositionsTracker")


class PositionsTracker(object):
    def __init__(self, data_frequency: typing.Literal["daily", "minute"]):
        if data_frequency not in {"daily", "minute"}:
            raise ValueError("data_frequency must be one of 'daily', 'minute'")
        self.data_frequency = data_frequency
        self.data: typing.OrderedDict[Asset, Position] = OrderedDict()
        self._unpaid_dividends = {}
        self._unpaid_stock_dividends = {}
        self._positions_store = {}

        self._need_update_stats = True  # ask chatgpt why use the word "dirty"
        self._stats = PositionStats()  # TODO: implement this

    def __getitem__(self, item: Asset):
        return self.data[item]

    def update_position(
            self,
            asset: Asset,
            amount: int = None,
            last_sale_price: float = None,
            last_sale_date: pd.Timestamp = None,
            cost_basis: float = None,
    ):
        self._need_update_stats = True

        if asset not in self.data:
            position = Position(asset)
            self.data[asset] = position

        else:
            position = self.data[asset]

        if amount is not None:
            position.amount = amount
        if last_sale_price is not None:
            position.last_sale_price = last_sale_price
        if last_sale_date is not None:
            position.last_sale_date = last_sale_date
        if cost_basis is not None:
            position.cost_basis = cost_basis

    def handle_transaction(self, txn: Transaction):
        self._need_update_stats = True

        asset = txn.asset

        if asset not in self.data:
            position = Position(asset)
            self.data[asset] = position
        else:
            position = self.data[asset]

        position.handle_transaction(txn)

    def handle_commission(self, asset: Asset, cost: float) -> None:
        if asset in self.data:
            self._need_update_stats = True
            self.data[asset].handle_commission(asset, cost)

    def handle_splits(self, splits: typing.List[typing.Tuple[Asset, float]]) -> float:
        """Process a list of splits by modifying any positions as needed
        Returns
        -------
        int: The leftover cash from fractional shares after modifying each position
        """
        total_leftover_cash = 0

        for asset, ratio in splits:
            if asset in self.data:
                self._need_update_stats = True

                position = self.data[asset]
                leftover_cash = position.handle_split(asset, ratio)
                total_leftover_cash += leftover_cash

        return total_leftover_cash

    def earn_dividends(self):
        # set _need_stats_update to True
        pass

    def pay_dividends(self):
        # set _need_stats_update to True
        pass

    def maybe_create_position_transaction(self):
        pass

    def get_positions(self) -> typing.Dict[Asset, Position]:
        # positions = self._positions_store
        #
        # for asset, position in self.positions.items():
        #     positions[asset] = position
        # TODO: 尽量只写
        return self.data

    def get_position(self, asset: Asset) -> Position:
        try:
            return self.data[asset]
        except KeyError:
            return Position(asset)

    def get_position_list(self) -> typing.List[Position]:
        return [
            pos for asset, pos in self.data.items() if pos.amount != 0
        ]

    def sync_last_sale_price(
            self,
            dt: pd.Timestamp,
            data_portal: DataPortal,
            handle_non_market_minutes=False,
    ):
        # 现在理解作用了吧
        self._need_update_stats = True

        if handle_non_market_minutes:
            raise NotImplementedError

        for position in self.data.values():
            position.last_sale_price = data_portal.get_scalar_asset_spot_value(
                asset=position.asset,
                filed="price",
                dt=dt,
                data_frequency=self.data_frequency,
            )
            position.last_sale_date = dt

    @property
    def stats(self):
        """The current status of the positions

        Returns
        -------
        stats: PositionStats
            The current stats position stats

        Notes
        -----
        This is cached, repeated access will not recompute the stats until
        the stats may have changed.
        """
        if self._need_update_stats:
            self._stats.calculate_positions_tracker_stats(self.data)
            self._need_update_stats = False
        return self._stats


class PositionStats:
    """Compute values from the current positions

    Attributes
    ----------
    """

    def __init__(self):
        self.gross_exposure = 0
        self.gross_exposure = 0
        self.gross_value = 0
        self.long_exposure = 0
        self.long_value = 0
        self.net_exposure = 0
        self.net_value = 0
        self.short_exposure = 0
        self.short_value = 0
        self.longs_count = 0
        self.shorts_count = 0

    def calculate_positions_tracker_stats(self, positions):
        long_value = 0.0
        short_value = 0.0
        long_exposure = 0.0
        short_exposure = 0.0
        longs_count = 0
        shorts_count = 0

        for position in positions.values():
            # NOTE: this loop does a lot of stuff!
            # we call this function every time the portfolio value is needed,
            # which is at least once per simulation day, so let's not iterate
            # through every single position multiple times.
            exposure = position.amount * position.last_sale_price

            if type(position.asset) is Future:
                # Futures don't have an inherent position value.
                value = 0

                # unchecked cast, this is safe because we do a type check above
                exposure *= position.asset.price_multiplier
            else:
                value = exposure

            if exposure > 0:
                longs_count += 1
                long_value += value
                long_exposure += exposure
            elif exposure < 0:
                shorts_count += 1
                short_value += value
                short_exposure += exposure

            # index[ix] = position.asset.sid
            # position_exposure[ix] = exposure
            # ix += 1

        net_value = long_value + short_value
        gross_value = long_value - short_value

        net_exposure = long_exposure + short_exposure
        gross_exposure = long_exposure - short_exposure

        self.gross_exposure = gross_exposure
        self.gross_value = gross_value
        self.long_exposure = long_exposure
        self.long_value = long_value
        self.longs_count = longs_count
        self.net_exposure = net_exposure
        self.net_value = net_value
        self.short_exposure = short_exposure
        self.short_value = short_value
        self.shorts_count = shorts_count














# Time Operation
from datetime import datetime
from dateutil.relativedelta import relativedelta
import time


class Timer(object):
    """
    >>> import numpy as np
    >>> with Timer("mytest.ipynb np.mean") as f:
    ...     x = np.random.normal(0, 1, (1000000, 100))
    ...     np.mean(x, axis=1)
    """
    def __init__(self, program: str = None):
        self.program = program
        self.start = None

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, exc_type, exc_value, trace):
        print("{}: {:.4f}ms".format(self.program, 1000*(time.time() - self.start)))


from torchqtm.base import BaseAlpha, Momentum, Reversion, Fundamental, Volatility, Parameter, Technical

__all__ = [
    "BaseAlpha",
    "Momentum",
    "Reversion",
    "Fundamental",
    "Volatility",
    "Technical",
    "Parameter",
]




import logging

import pandas as pd

log = logging.getLogger("SimulationParameters")

DEFAULT_CAPITAL_BASE = 1e5
from torchqtm.utils.exchange_calendar import XSHGExchangeCalendar
from torchqtm.types import DATA_FREQUENCIES
from functools import lru_cache


class SimulationParameters(object):
    """
    The collections of parameters for creating backtesting environments.
    """
    def __init__(
            self,
            start_session: pd.Timestamp,
            end_session: pd.Timestamp,
            trading_calendar: XSHGExchangeCalendar,
            capital_base: float = DEFAULT_CAPITAL_BASE,
            emission_rate: DATA_FREQUENCIES = "daily",
            data_frequency: DATA_FREQUENCIES = "daily",
    ):
        assert type(start_session) == pd.Timestamp
        assert type(end_session) == pd.Timestamp

        assert trading_calendar is not None, "Must pass in trading calendar!"
        assert start_session <= end_session, "Period start falls after period end."

        assert start_session >= trading_calendar.first_session
        assert end_session <= trading_calendar.last_session

        self._start_session = start_session
        self._end_session = end_session

        self._capital_base = capital_base

        self._emission_rate = emission_rate
        self._data_frequency = data_frequency

        self._trading_calendar = trading_calendar

        # if the start_date is not the trading_session(Almost never happen)

        self._start_session = trading_calendar.minute_to_session(self._start_session)

        self._end_session = trading_calendar.minute_to_session(self._end_session)

    @property
    def capital_base(self):
        return self._capital_base

    @property
    def emission_rate(self):
        return self._emission_rate

    @property
    def data_frequency(self):
        return self._data_frequency

    def set_data_frequency(self, val):
        self._data_frequency = val

    @property
    def start_session(self):
        return self._start_session

    @property
    def end_session(self):
        return self._end_session

    @property
    def trading_calendar(self):
        return self._trading_calendar

    def create_new(self, start_session, end_session, data_frequency=None):
        raise NotImplementedError

    @property
    @lru_cache(None)
    def sessions(self):
        return self._trading_calendar.sessions_in_range(self._start_session, self._end_session)

    def __repr__(self):
        template = """
{class_name}(
    start_session={start_session},
    end_session={end_session},
    capital_base={capital_base},
    data_frequency={data_frequency},
    emission_rate={emission_rate},
    trading_calendar={trading_calendar},
)
""".strip()
        return template.format(
            class_name=self.__class__.__name__,
            start_session=self.start_session,
            end_session=self.end_session,
            capital_base=self.capital_base,
            data_frequency=self.data_frequency,
            emission_rate=self.emission_rate,
            trading_calendar=self.trading_calendar,
        )


if __name__ == "__main__":
    start_date = pd.Timestamp("20100101")
    end_date = pd.Timestamp("20100201")
    from torchqtm.utils.calendar_utils import get_calendar
    test = SimulationParameters(start_date, end_date, get_calendar("XSHG"))
    test.sessions
    repr(test)
    print(test)
    print(test.start_session)
    print(test.end_session)







import pandas as pd
import torch
import torchqtm.op.functional as F
getattr(F, "ts_mean")

F['ts_mean']
from torchqtm.utils import Timer
import numpy as np
import pandas as pd
a = np.random.normal(0, 1, (252 * 10, 50000))
b = np.random.normal(0, 1, (252 * 100, 500))
c = 2*b + 1


with Timer():
    a = a.astype(np.float64)

am = pd.DataFrame(a)

with Timer():
    am.rolling(20).corr(am)

with Timer():
    F.ts_corr(a, a, 20)


F.ts_corr(a, a, 20)

am.rolling(20).mean().tail(5)

F.ts_mean(a, 20)[-5:]




from torchqtm.types import ASSET_TYPE


class Asset:
    type: ASSET_TYPE = ...

    def __init__(self,
                 symbol: str,  # For example "000001.XSHE"
                 asset_name: str = None,
                 start_date: int = None,
                 first_traded: int = None,
                 tick_size: float = 0.01):

        self.symbol: str = symbol
        self.asset_name: str = asset_name
        self.start_date: int = start_date  # Numerical-based
        self.first_traded: int = first_traded  # Numerical-based
        self.tick_size: float = tick_size
        self.sid: int = self.__int__()
        self.price_multiplier: float = 1.

    def __int__(self):
        number_string = self.symbol.split(".")[0]
        # convert the string to an integer
        number = int(number_string)
        return number

    def __repr__(self):
        if self.symbol:
            return f"<{type(self).__name__}({self.symbol})>"

    def __eq__(self, other):
        return self.symbol == other.symbol

    # TODO: implement it
    def is_exchange_open(self):
        pass

    def __hash__(self):
        return hash(self.symbol)


class Equity(Asset):
    type = ASSET_TYPE.Equity

    @classmethod
    def from_int(cls, int_id):
        raise NotImplementedError


class Future(Asset):
    type = ASSET_TYPE.Future


class Index(Asset):
    type = ASSET_TYPE.Index


class Fund(Asset):
    type = ASSET_TYPE.Fund




import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.op.functional import *
import pandas as pd
import numpy as np
from abc import ABCMeta, abstractmethod
from torchqtm.tdbt.backtest import BackTestEnv
from torchqtm.base import BaseAlpha


# __all__ = [f"Alpha{str(i).zfill(3)}" for i in range(1, 102)]
# https://github1s.com/yli188/WorldQuant_alpha101_code/blob/master/101Alpha_code_1.py#L378
# https://chat.openai.com/share/7147e5ec-d1f9-4282-9394-c59fd35ee13b
class WQAlpha101(BaseAlpha, metaclass=ABCMeta):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'WorldQuant Alpha101'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Alpha001(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * ts_rank(cs_rank(self.low), 9)
        return self.data


class Alpha002(WQAlpha101):
    """
     (-1 * correlation(rank(delta(log(volume), 2)), rank(((close - open) / open)), 6))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * ts_corr(cs_rank(ts_delta(log(self.volume), 2)), cs_rank((self.close - self.open) / self.open),
                                 6)
        return self.data


class Alpha003(WQAlpha101):
    """
    The Spearman correlation of open and volume
    -1 * correlation(rank(open), rank(volume), 10)
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * ts_corr(cs_rank(self.open), cs_rank(self.volume), 10)
        return self.data


class Alpha004(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * ts_rank(cs_rank(self.low), 9)
        return self.data


class Alpha005(WQAlpha101):
    """
     (rank((open - (sum(vwap, 10) / 10))) * (-1 * abs(rank((close - vwap)))))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = (cs_rank((self.open - ts_mean(self.vwap, 10))) * (-1 * abs(cs_rank((self.close - self.vwap)))))
        return self.data


class Alpha006(WQAlpha101):
    """
     (-1 * correlation(open, volume, 10))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha007(WQAlpha101):
    """
    ((adv20 < volume) ? ((-1 * ts_rank(abs(delta(close, 7)), 60)) * sign(delta(close, 7))) : (-1* 1))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        condition = self.adv20 < self.volume
        value_if_true = -1 * ts_rank(abs(ts_delta(self.close, 7)), 60) * sign(ts_delta(self.close, 7))
        value_if_false = -1
        self.data = if_else(condition, value_if_true, value_if_false)
        return self.data


class Alpha008(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * (cs_rank(((ts_sum(self.open, 5) * ts_sum(self.returns, 5)) -
                                   ts_delay((ts_sum(self.open, 5) * ts_sum(self.returns, 5)), 10))))
        return self.data

    def __repr__(self):
        return "Alpha: (-1 * rank(((sum(open, 5) * sum(returns, 5)) - delay((sum(open, 5) * sum(returns, 5)),10))))"


class Alpha009(WQAlpha101):
    """
     (-1 * correlation(rank(delta(log(volume), 2)), rank(((close - open) / open)), 6))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = -1 * ts_corr(cs_rank(ts_delta(log(self.volume), 2)), cs_rank((self.close - self.open) / self.open),
                                 6)
        return self.data

    def __repr__(self):
        return "(-1 * correlation(rank(delta(log(volume), 2)), rank(((close - open) / open)), 6))"


class Alpha010(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        pass

    def forward(self):
        return self.data


class Alpha011(WQAlpha101):
    """
    ((rank(ts_max((vwap - close), 3)) + rank(ts_min((vwap - close), 3))) *rank(delta(volume, 3)))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = ((F.cs_rank(F.ts_max((self.env.Vwap - self.env.Close), 3)) +
                      F.cs_rank(F.ts_min((self.env.Vwap - self.env.Close), 3))) * F.cs_rank(
            F.ts_delta(self.env.Volume, 3)))
        return self.data

    def __repr__(self):
        return "((rank(ts_max((vwap - close), 3)) + rank(ts_min((vwap - close), 3))) *rank(delta(volume, 3)))"


class Alpha012(WQAlpha101):
    """
    (sign(delta(volume, 1)) * (-1 * delta(close, 1)))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = sign(ts_delta(self.volume, 1)) * (-1 * ts_delta(self.close, 1))
        return self.data

    def __repr__(self):
        return "(sign(delta(volume, 1)) * (-1 * delta(close, 1)))"


# TODO: implement it
class Alpha013(WQAlpha101):
    """
    (-1 * rank(covariance(rank(close), rank(volume), 5)))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        # self.data = (-1 * cs_rank(covariance(rank(close), rank(volume), 5)))
        # return self.data
        pass

    def __repr__(self):
        return "(sign(delta(volume, 1)) * (-1 * delta(close, 1)))"


class Alpha014(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha015(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha016(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha017(WQAlpha101):
    """
     (((-1 * rank(ts_rank(close, 10))) * rank(delta(delta(close, 1), 1))) *rank(ts_rank((volume / adv20), 5)))
    """

    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        adv20 = ts_mean(self.env.Volume, 20)
        self.data = -1 * (cs_rank(ts_rank(self.env.Close, 10)) *
                          cs_rank(ts_delta(ts_delta(self.env.Close, 1), 1)) *
                          cs_rank(ts_rank((self.env.Volume / adv20), 5)))
        return self.data


class Alpha018(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * rank(((stddev(abs((close - open)), 5) + (close - open)) + correlation(close, open,10))))"

    def forward(self):
        df = ts_corr(self.close, self.open, 10)
        self.data = -1 * (cs_rank((ts_std_dev(abs(self.close - self.open), 5) + (self.close - self.open)) +
                                  df))
        return self.data


class Alpha019(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * sign(((close - delay(close, 7)) + delta(close, 7)))) * (1 + rank((1 + sum(returns,250)))))"

    def forward(self):
        self.data = ((-1 * sign((self.close - ts_delay(self.close, 7)) + ts_delta(self.close, 7))) *
                     (1 + cs_rank(1 + ts_sum(self.returns, 250))))
        return self.data


class Alpha020(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(((-1 * rank((open - delay(high, 1)))) * rank((open - delay(close, 1)))) * rank((open -delay(low, 1))))"

    def forward(self):
        self.data = -1 * (cs_rank(self.open - ts_delay(self.high, 1)) *
                          cs_rank(self.open - ts_delay(self.close, 1)) *
                          cs_rank(self.open - ts_delay(self.low, 1)))
        return self.data


# class Alpha021(WQAlpha101):
#     def __init__(self, env):
#         super().__init__(env)
#
#     def __repr__(self):
#         return "((((sum(close, 8) / 8) + stddev(close, 8)) < (sum(close, 2) / 2)) ? (-1 * 1) : (((sum(close," \
#                "2) / 2) < ((sum(close, 8) / 8) - stddev(close, 8))) ? 1 : (((1 < (volume / adv20)) || ((volume " \
#                "/adv20) == 1)) ? 1 : (-1 * 1))))"
#
#     def forward(self):
#         cond_1 = ts_mean(self.close, 8) + ts_std_dev(self.close, 8) < ts_mean(self.close, 2)
#         cond_2 = ts_mean(self.volume, 20) / self.volume < 1
#         condition =
#         self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
#         return self.data


class Alpha022(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * (delta(correlation(high, volume, 5), 5) * rank(stddev(close, 20))))"

    def forward(self):
        df = ts_corr(self.high, self.volume, 5)
        self.data = -1 * ts_delta(df, 5) * cs_rank(ts_std_dev(self.close, 20))
        return self.data


class Alpha023(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(((sum(high, 20) / 20) < high) ? (-1 * delta(high, 2)) : 0)"

    def forward(self):
        condition = less(ts_mean(self.high, 20), self.high)
        value_if_true = -1 * ts_delta(self.high, 2)
        value_if_false = 0
        self.data = if_else(condition, value_if_true, value_if_false)
        return self.data


class Alpha024(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((((delta((sum(close, 100) / 100), 100) / delay(close, 100)) < 0.05) ||((delta((sum(close, " \
               "100) / 100), 100) / delay(close, 100)) == 0.05)) ? (-1 * (close - ts_min(close,100))) : (-1 * delta(" \
               "close, 3)))"

    def forward(self):
        condition = ts_delta(ts_mean(self.close, 100), 100) / geq(ts_delay(self.close, 100), 0.05)
        value_if_true = -1 * (self.close - ts_min(self.close, 100))
        value_if_false = -1 * ts_delta(self.close, 3)
        self.data = if_else(condition, value_if_true, value_if_false)
        return self.data


class Alpha025(WQAlpha101):
    """
               [rank]
                 |
               [mul]
               /   \
           [mul]   [sub]
           /   \     /   \
        [mul] vwap high close
         /   \
       [mul] adv20
       /   \
      -1 returns

    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "rank(((((-1 * returns) * adv20) * vwap) * (high - close)))"

    def forward(self):
        self.data = cs_rank(((((-1 * self.returns) * self.adv20) * self.vwap) * (self.high - self.close)))
        return self.data


class Alpha026(WQAlpha101):
    """
        [mul]
        /   \
      -1  [ts_max]
          /   \
    [correlation] 3
     /   |   \
    [ts_rank] [ts_rank] 5
     /   \    /   \
    volume 5 high  5

    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * ts_max(correlation(ts_rank(volume, 5), ts_rank(high, 5), 5), 3))"

    def forward(self):
        df = ts_corr(ts_rank(self.volume, 5), ts_rank(self.high, 5), 5)
        self.data = -1 * ts_max(df, 3)
        return self.data


class Alpha027(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((0.5 < rank((sum(correlation(rank(volume), rank(vwap), 6), 2) / 2.0))) ? (-1 * 1) : 1)"

    def forward(self):
        condition = (0.5 < cs_rank((ts_mean(ts_corr(cs_rank(self.volume), cs_rank(self.vwap), 6), 2))))
        value_if_true = -1
        value_if_false = 1
        self.data = if_else(condition, value_if_true, value_if_false)
        return self.data


# class Alpha028(WQAlpha101):
#     def __init__(self, env):
#         super().__init__(env)
#
#     def __repr__(self):
#         return "scale(((correlation(adv20, low, 5) + ((high + low) / 2)) - close))"
#
#     def forward(self):
#         self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
#         return self.data


# class Alpha029(WQAlpha101):
#     def __init__(self, env):
#         super().__init__(env)
#
#     def __repr__(self):
#         return "(min(product(rank(rank(scale(log(sum(ts_min(rank(rank((-1 * rank(delta((close - 1),5))))), 2), " \
#                "1))))), 1), 5) + ts_rank(delay((-1 * returns), 6), 5))"
#
#     def forward(self):
#         self.data = (ts_min(cs_rank(cs_rank(scale(log(ts_sum(rank(rank(-1 * rank(delta((self.close - 1), 5)))), 2))))), 5) +
#                 ts_rank(delay((-1 * self.returns), 6), 5))
#         return self.data


class Alpha030(WQAlpha101):
    """
                   [div]
                   /   \
               [mul]   [sum]
               /   \       /   \
          [sub]    [sum] volume 20
          /   \       /   \
        1.0 [rank] volume 5
             |
           [add]
           /   \
       [add]  [sign]
       /   \    |
    [sign] [sign] [delay]
     /   \   /   \  /   \
    [delay] [delay] close 3
     /   \    /   \
    close 1 close 2
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(((1.0 - rank(((sign((close - delay(close, 1))) + sign((delay(close, 1) - delay(close, 2)))) +sign((" \
               "delay(close, 2) - delay(close, 3)))))) * sum(volume, 5)) / sum(volume, 20))"

    def forward(self):
        delta_close = ts_delta(self.close, 1)
        inner = sign(delta_close) + sign(ts_delay(delta_close, 1)) + sign(ts_delay(delta_close, 2))
        self.data = ((1.0 - cs_rank(inner)) * ts_sum(self.volume, 5)) / ts_sum(self.volume, 20)
        return self.data


# TODO: implement it
class Alpha031(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "((rank(rank(rank(decay_linear((-1 * rank(rank(delta(close, 10)))), 10)))) + " \
               "rank((-1 *delta(close, 3)))) + sign(scale(correlation(adv20, low, 12))))"

    def forward(self):
        pass


# TODO: implement it
class Alpha032(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "(scale(((sum(close, 7) / 7) - close)) + (20 * scale(correlation(vwap, delay(close, 5),230))))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha033(WQAlpha101):
    """
             [rank]
             |
            [mul]
            /   \
          -1    [pow]
                /   \
             [sub]   1
             /   \
            1   [div]
                /   \
             open close
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "rank((-1 * ((1 - (open / close))^1)))"

    def forward(self):
        self.data = cs_rank(-1 + (self.open / self.close))
        return self.data


class Alpha034(WQAlpha101):
    """
             [rank]
             |
           [add]
           /   \
       [sub]   [sub]
       /   \     /   \
      1  [rank] 1  [rank]
         |         |
      [div]    [delta]
     /   \     /   \
    [stddev] [stddev] close 1
     /   \     /   \
    returns 2 returns 5

    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "rank(((1 - rank((stddev(returns, 2) / stddev(returns, 5)))) + (1 - rank(delta(close, 1)))))"

    def forward(self):
        temp = ts_std_dev(self.returns, 2) / ts_std_dev(self.returns, 5)
        self.data = cs_rank(2 - cs_rank(temp) - cs_rank(ts_delta(self.close, 1)))
        return self.data


class Alpha035(WQAlpha101):
    """
              [mul]
              /   \
           [mul]  [sub]
           /   \     / \
     [ts_Rank] [sub] 1 [ts_Rank]
     /   \     /     /   \
    volume 32 1 [ts_Rank] returns 32
               | /   \
             [sub]   16
             / | \
         [add] low
         /   \
       close high
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((ts_Rank(volume, 32) * (1 - Ts_Rank(((close + high) - low), 16))) * (1 -Ts_Rank(returns, 32)))"

    def forward(self):
        self.data = ((ts_rank(self.volume, 32) *
                      (1 - ts_rank(self.close + self.high - self.low, 16))) *
                     (1 - ts_rank(self.returns, 32)))
        return self.data


class Alpha036(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False  # It seems too complex and use too long data.

    def __repr__(self):
        return "(((((2.21 * rank(correlation((close - open), delay(volume, 1), 15))) + (0.7 * rank((open- close)))) + " \
               "(0.73 * rank(Ts_Rank(delay((-1 * returns), 6), 5)))) + rank(abs(correlation(vwap,adv20, 6)))) + (0.6 " \
               "* rank((((sum(close, 200) / 200) - open) * (close - open)))))"

    def forward(self):
        pass


class Alpha037(WQAlpha101):
    """
              [add]
              /   \
        [rank]   [rank]
        /          |
    [correlation] [sub]
     /   |   \     /   \
    [delay] close 200 open close
     /   \
    [sub] 1
     /   \
    open close
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(rank(correlation(delay((open - close), 1), close, 200)) + rank((open - close)))"

    def forward(self):
        self.data = cs_rank(ts_corr(ts_delay(self.open - self.close, 1), self.close, 200)) \
                    + cs_rank(self.open - self.close)
        return self.data


class Alpha038(WQAlpha101):
    """
              (*)
              / \
             /   \
            /     \
        [mul]    [cs_rank]
         /  \        |
        -1  [cs_rank] [divide]
            |        /   \
         [ts_rank] self.close self.open
            |
         [self.close, 10]
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(Ts_Rank(close, 10))) * rank((close / open)))"

    def forward(self):
        self.data = (-1 * cs_rank(ts_rank(self.close, 10))) * cs_rank(divide(self.close, self.open))
        return self.data


class Alpha039(WQAlpha101):
    """
            [mul]
            /   \
         [mul]  [add]
         /   \     / \
       -1  [rank] 1 [rank]
            |         |
       [mul]     [sum]
       /   \         |
    [delta] [sub] [returns, 250]
     /   \   /   \
    close 7  1  [rank]
               |
          [decay_linear]
           /        \
      [div]         9
     /   \
    volume adv20
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank((delta(close, 7) * (1 - rank(decay_linear((volume / adv20), 9)))))) * (1 +rank(sum(" \
               "returns, 250))))"

    def forward(self):
        self.data = ((-1 * cs_rank(
            ts_delta(self.close, 7) * (1 - cs_rank(ts_decay_linear((self.volume / self.adv20), 9).CLOSE)))) *
                     (1 + cs_rank(ts_mean(self.returns, 250))))
        return self.data


class Alpha040(WQAlpha101):
    """
            [mul]
            /   \
        [mul]  [correlation]
        /   \     /   |   \
      -1  [rank] high volume 10
          |
       [stddev]
       /   \
     high  10
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_std_dev(self.high, 10)) * ts_corr(self.high, self.volume, 10)
        return self.data


class Alpha041(WQAlpha101):
    """
           [sub]
           /   \
       [pow]   vwap
       /   \
     [mul]  0.5
     /   \
    high low
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(((high * low)^0.5) - vwap)"

    def forward(self):
        self.data = pow((self.high * self.low), 0.5) - self.vwap
        return self.data


class Alpha042(WQAlpha101):
    """
        [div]
        /   \
    [rank]  [rank]
      |       |
    [sub]   [add]
    /   \   /   \
    vwap close vwap close
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(rank((vwap - close)) / rank((vwap + close)))"

    def forward(self):
        self.data = cs_rank((self.vwap - self.close)) / cs_rank((self.vwap + self.close))
        return self.data


class Alpha043(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "ts_rank(self.volume / adv20, 20) * ts_rank((-1 * delta(self.close, 7)), 8)"

    def forward(self):
        self.data = ts_rank(self.volume / self.adv20, 20) * ts_rank((-1 * ts_delta(self.close, 7)), 8)
        return self.data


class Alpha044(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * correlation(high, rank(volume), 5))"

    def forward(self):
        self.data = ts_corr(self.high, cs_rank(self.volume), 5)
        return self.data


class Alpha045(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * ((rank((sum(delay(close, 5), 20) / 20)) * correlation(close, volume, 2)) *rank(correlation(sum(" \
               "close, 5), sum(close, 20), 2))))"

    def forward(self):
        df = ts_corr(self.close, self.volume, 2)
        self.data = -1 * (cs_rank(ts_mean(ts_delay(self.close, 5), 20)) * df *
                          cs_rank(ts_corr(ts_sum(self.close, 5), ts_sum(self.close, 20), 2)))
        return self.data


class Alpha046(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "((0.25 < (((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10))) ?(-1 * " \
               "1) : (((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < 0) ? 1 :(" \
               "(-1 * 1) * (close - delay(close, 1)))))"

    def forward(self):
        # inner = ((ts_delay(self.close, 20) - ts_delay(self.close, 10)) / 10) - ((ts_delay(self.close, 10) - self.close) / 10)
        # alpha = (-1 * ts_delta(self.close))
        # alpha[inner < 0] = 1
        # alpha[inner > 0.25] = -1
        # self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha047(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((((rank((1 / close)) * volume) / adv20) * ((high * rank((high - close))) / (sum(high, 5) /5))) - " \
               "rank((vwap - delay(vwap, 5))))"

    def forward(self):
        self.data = ((((cs_rank((1 / self.close)) * self.volume) / self.adv20) *
                      ((self.high * cs_rank((self.high - self.close))) / (ts_mean(self.high, 5) / 5)))
                     - cs_rank((self.vwap - ts_delay(self.vwap, 5))))
        return self.data


class Alpha048(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "(indneutralize(((correlation(delta(close, 1), delta(delay(close, 1), 1), 250) *delta(close, " \
               "1)) / close), IndClass.subindustry) / sum(((delta(close, 1) / delay(close, 1))^2), 250))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha049(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < (-1 *0.1)) ? 1 " \
               ": ((-1 * 1) * (close - delay(close, 1))))"

    def forward(self):
        condition = (((ts_delay(self.close, 20) - ts_delay(self.close, 10)) / 10) - (
                (ts_delay(self.close, 10) - self.close) / 10)) < -0.1
        value_if_true = 1
        value_if_false = -1 * ts_delta(self.close, 1)
        self.data = if_else(condition, value_if_true, value_if_false)
        return self.data


class Alpha050(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * ts_max(rank(correlation(rank(volume), rank(vwap), 5)), 5))"

    def forward(self):
        self.data = (-1 * ts_max(cs_rank(ts_corr(cs_rank(self.volume), cs_rank(self.vwap), 5)), 5))
        return self.data


class Alpha051(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < (-1 *0.05)) ? " \
               "1 : ((-1 * 1) * (close - delay(close, 1))))"

    def forward(self):
        condition = (((ts_delay(self.close, 20) - ts_delay(self.close, 10)) / 10) - (
                (ts_delay(self.close, 10) - self.close) / 10)) < -0.05
        value_if_true = 1
        value_if_false = -1 * ts_delta(self.close, 1)
        self.data = if_else(condition, value_if_true, value_if_false)
        return self.data


class Alpha052(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((((-1 * ts_min(low, 5)) + delay(ts_min(low, 5), 5)) * rank(((sum(returns, 240) -sum(returns, " \
               "20)) / 220))) * ts_rank(volume, 5))"

    def forward(self):
        self.data = (((-1 * ts_delta(ts_min(self.low, 5), 5)) *
                      cs_rank(((ts_sum(self.returns, 240) - ts_sum(self.returns, 20)) / 220))) * ts_rank(self.volume,
                                                                                                         5))
        return self.data


class Alpha053(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "(-1 * delta((((close - low) - (high - close)) / (close - low)), 9))"

    def forward(self):
        inner = (self.close - self.low)
        self.data = -1 * ts_delta((((self.close - self.low) - (self.high - self.close)) / inner), 9)
        return self.data


# Alpha#54	 ((-1 * ((low - close) * (open^5))) / ((low - high) * (close^5)))
def alpha054(self):
    inner = (self.low - self.high).replace(0, -0.0001)
    return -1 * (self.low - self.close) * (self.open ** 5) / (inner * (self.close ** 5))


class Alpha054(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * ((low - close) * (open^5))) / ((low - high) * (close^5)))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha055(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(-1 * correlation(rank(((close - ts_min(low, 12)) / (ts_max(high, 12) - ts_min(low,12)))), " \
               "rank(volume), 6))"

    def forward(self):
        divisor = (ts_max(self.high, 12) - ts_min(self.low, 12)).replace(0, 0.0001)
        inner = (self.close - ts_min(self.low, 12)) / divisor
        self.data = ts_corr(cs_rank(inner), cs_rank(self.volume), 6)
        return self.data


class Alpha056(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "(0 - (1 * (rank((sum(returns, 10) / sum(sum(returns, 2), 3))) * rank((returns * cap)))))"

    def forward(self):
        return self.data


class Alpha057(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = (self.close - self.vwap) / ts_decay_linear(cs_rank(ts_arg_max(self.close, 30)), 2)
        return self.data

# Alpha#58	 (-1 * Ts_Rank(decay_linear(correlation(IndNeutralize(vwap, IndClass.sector), volume,3.92795), 7.89291), 5.50322))

# Alpha#59	 (-1 * Ts_Rank(decay_linear(correlation(IndNeutralize(((vwap * 0.728317) + (vwap *(1 - 0.728317))), IndClass.industry), volume, 4.25197), 16.2289), 8.19648))


class Alpha060(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "(0 - (1 * ((2 * scale(rank(((((close - low) - (high - close)) / (high - low)) * volume)))) -scale(" \
               "rank(ts_argmax(close, 10))))))"

    def forward(self):
        divisor = (self.high - self.low) + 0.001
        inner = ((self.close - self.low) - (self.high - self.close)) * self.volume / divisor
        # self.data = - ((2 * ts_scale(cs_rank(inner))) - scale(rank(ts_argmax(self.close, 10))))
        return self.data


class Alpha061(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(rank((vwap - ts_min(vwap, 16.1219))) < rank(correlation(vwap, adv180, 17.9282)))"

    def forward(self):
        adv180 = ts_mean(self.volume, 180)
        self.data = (cs_rank((self.vwap - ts_min(self.vwap, 16))) < cs_rank(ts_corr(self.vwap, adv180, 18)))
        return self.data


class Alpha062(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((rank(correlation(vwap, sum(adv20, 22.4101), 9.91009)) < rank(((rank(open) +rank(open)) < (rank(((" \
               "high + low) / 2)) + rank(high))))) * -1)"

    def forward(self):
        adv20 = ts_mean(self.volume, 20)
        self.data = ((cs_rank(ts_corr(self.vwap, ts_mean(adv20, 22), 10)) < cs_rank(
            ((cs_rank(self.open) + cs_rank(self.open)) < (cs_rank(((self.high + self.low) / 2)) + cs_rank(self.high))))) * -1)
        return self.data


# Alpha#63	 ((rank(decay_linear(delta(IndNeutralize(close, IndClass.industry), 2.25164), 8.22237))- rank(decay_linear(correlation(((vwap * 0.318108) + (open * (1 - 0.318108))), sum(adv180,37.2467), 13.557), 12.2883))) * -1)
class Alpha063(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha064(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((rank(correlation(sum(((open * 0.178404) + (low * (1 - 0.178404))), 12.7054),sum(adv120, 12.7054), " \
               "16.6208)) < rank(delta(((((high + low) / 2) * 0.178404) + (vwap * (1 -0.178404))), 3.69741))) * -1)"

    def forward(self):
        adv120 = ts_mean(self.volume, 120)
        self.data = ((cs_rank(
            ts_corr(ts_mean(((self.open * 0.178404) + (self.low * (1 - 0.178404))), 13), ts_mean(adv120, 13), 17)) < cs_rank(
            ts_delta(((((self.high + self.low) / 2) * 0.178404) + (self.vwap * (1 - 0.178404))), 3.69741))) * -1)
        return self.data


class Alpha065(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((rank(correlation(((open * 0.00817205) + (vwap * (1 - 0.00817205))), sum(adv60,8.6911), 6.40374)) < " \
               "rank((open - ts_min(open, 13.635)))) * -1)"

    def forward(self):
        adv60 = ts_mean(self.volume, 60)
        self.data = ((cs_rank(
            ts_corr(((self.open * 0.00817205) + (self.vwap * (1 - 0.00817205))), ts_mean(adv60, 9), 6)) < cs_rank(
            (self.open - ts_min(self.open, 14)))) * -1)
        return self.data


# # Alpha#66	 ((rank(decay_linear(delta(vwap, 3.51013), 7.23052)) + Ts_Rank(decay_linear(((((low* 0.96633) + (low * (1 - 0.96633))) - vwap) / (open - ((high + low) / 2))), 11.4157), 6.72611)) * -1)
# def alpha066(self):
#     return ((rank(decay_linear(delta(self.vwap, 4).to_frame(), 7).CLOSE) + ts_rank(decay_linear(((((
#                                                                                                            self.low * 0.96633) + (
#                                                                                                            self.low * (
#                                                                                                            1 - 0.96633))) - self.vwap) / (
#                                                                                                          self.open - (
#                                                                                                          (
#                                                                                                                  self.high + self.low) / 2))).to_frame(),
#                                                                                                 11).CLOSE, 7)) * -1)


# Alpha#67	 ((rank((high - ts_min(high, 2.14593)))^rank(correlation(IndNeutralize(vwap,IndClass.sector), IndNeutralize(adv20, IndClass.subindustry), 6.02936))) * -1)
class Alpha067(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False # 无数据

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha068(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((Ts_Rank(correlation(rank(high), rank(adv15), 8.91644), 13.9333) <rank(delta(((close * 0.518371) + (" \
               "low * (1 - 0.518371))), 1.06157))) * -1)"

    def forward(self):
        adv15 = ts_mean(self.volume, 15)
        self.data = ((ts_rank(ts_corr(cs_rank(self.high), cs_rank(adv15), 9), 14) < cs_rank(
            ts_delta(((self.close * 0.518371) + (self.low * (1 - 0.518371))), 1.06157))) * -1)
        return self.data



# Alpha#69	 ((rank(ts_max(delta(IndNeutralize(vwap, IndClass.industry), 2.72412),4.79344))^Ts_Rank(correlation(((close * 0.490655) + (vwap * (1 - 0.490655))), adv20, 4.92416),9.0615)) * -1)

# Alpha#70	 ((rank(delta(vwap, 1.29456))^Ts_Rank(correlation(IndNeutralize(close,IndClass.industry), adv50, 17.8256), 17.9171)) * -1)

# Alpha#71	 max(Ts_Rank(decay_linear(correlation(Ts_Rank(close, 3.43976), Ts_Rank(adv180,12.0647), 18.0175), 4.20501), 15.6948), Ts_Rank(decay_linear((rank(((low + open) - (vwap +vwap)))^2), 16.4662), 4.4388))
def alpha071(self):
    adv180 = sma(self.volume, 180)
    p1 = ts_rank(decay_linear(correlation(ts_rank(self.close, 3), ts_rank(adv180, 12), 18).to_frame(), 4).CLOSE, 16)
    p2 = ts_rank(
        decay_linear((rank(((self.low + self.open) - (self.vwap + self.vwap))).pow(2)).to_frame(), 16).CLOSE, 4)
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    df.at[df['p1'] >= df['p2'], 'max'] = df['p1']
    df.at[df['p2'] >= df['p1'], 'max'] = df['p2']
    return df['max']
    # return max(ts_rank(decay_linear(correlation(ts_rank(self.close, 3), ts_rank(adv180,12), 18).to_frame(), 4).CLOSE, 16), ts_rank(decay_linear((rank(((self.low + self.open) - (self.vwap +self.vwap))).pow(2)).to_frame(), 16).CLOSE, 4))


class Alpha071(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = False

    def __repr__(self):
        return "max(Ts_Rank(decay_linear(correlation(Ts_Rank(close, 3.43976), Ts_Rank(adv180,12.0647), 18.0175), " \
               "4.20501), 15.6948), Ts_Rank(decay_linear((rank(((low + open) - (vwap +vwap)))^2), 16.4662), 4.4388))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data


class Alpha072(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "(rank(decay_linear(correlation(((high + low) / 2), adv40, 8.93345), 10.1519)) /rank(decay_linear(" \
               "correlation(Ts_Rank(vwap, 3.72469), Ts_Rank(volume, 18.5188), 6.86671),2.95011)))"

    def forward(self):
        adv40 = ts_mean(self.volume, 40)
        numerator = cs_rank(ts_decay_linear(ts_corr((self.high + self.low)/2, adv40, 9), 10))
        denominator = cs_rank(ts_decay_linear(ts_corr(ts_rank(self.vwap, 4), ts_rank(self.volume, 19), 7), 3))
        self.data = numerator / denominator
        return self.data


# Alpha#73	 (max(rank(decay_linear(delta(vwap, 4.72775), 2.91864)),Ts_Rank(decay_linear(((delta(((open * 0.147155) + (low * (1 - 0.147155))), 2.03608) / ((open *0.147155) + (low * (1 - 0.147155)))) * -1), 3.33829), 16.7411)) * -1)
def alpha073(self):
    p1 = rank(decay_linear(delta(self.vwap, 5).to_frame(), 3).CLOSE)
    p2 = ts_rank(decay_linear(((delta(((self.open * 0.147155) + (self.low * (1 - 0.147155))), 2) / (
            (self.open * 0.147155) + (self.low * (1 - 0.147155)))) * -1).to_frame(), 3).CLOSE, 17)
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    df.at[df['p1'] >= df['p2'], 'max'] = df['p1']
    df.at[df['p2'] >= df['p1'], 'max'] = df['p2']
    return -1 * df['max']
    # return (max(rank(decay_linear(delta(self.vwap, 5).to_frame(), 3).CLOSE),ts_rank(decay_linear(((delta(((self.open * 0.147155) + (self.low * (1 - 0.147155))), 2) / ((self.open *0.147155) + (self.low * (1 - 0.147155)))) * -1).to_frame(), 3).CLOSE, 17)) * -1)

class Alpha073(WQAlpha101):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"

    def forward(self):
        self.data = -1 * cs_rank(ts_delta(self.returns, 3)) * ts_corr(self.open, self.volume, 10)
        return self.data
# Alpha#74	 ((rank(correlation(close, sum(adv30, 37.4843), 15.1365)) <rank(correlation(rank(((high * 0.0261661) + (vwap * (1 - 0.0261661)))), rank(volume), 11.4791)))* -1)
def alpha074(self):
    adv30 = sma(self.volume, 30)
    return ((rank(correlation(self.close, sma(adv30, 37), 15)) < rank(
        correlation(rank(((self.high * 0.0261661) + (self.vwap * (1 - 0.0261661)))), rank(self.volume), 11))) * -1)


# Alpha#75	 (rank(correlation(vwap, volume, 4.24304)) < rank(correlation(rank(low), rank(adv50),12.4413)))
def alpha075(self):
    adv50 = sma(self.volume, 50)
    return (rank(correlation(self.vwap, self.volume, 4)) < rank(correlation(rank(self.low), rank(adv50), 12)))


# Alpha#76	 (max(rank(decay_linear(delta(vwap, 1.24383), 11.8259)),Ts_Rank(decay_linear(Ts_Rank(correlation(IndNeutralize(low, IndClass.sector), adv81,8.14941), 19.569), 17.1543), 19.383)) * -1)

# Alpha#77	 min(rank(decay_linear(((((high + low) / 2) + high) - (vwap + high)), 20.0451)),rank(decay_linear(correlation(((high + low) / 2), adv40, 3.1614), 5.64125)))
def alpha077(self):
    adv40 = sma(self.volume, 40)
    p1 = rank(
        decay_linear(((((self.high + self.low) / 2) + self.high) - (self.vwap + self.high)).to_frame(), 20).CLOSE)
    p2 = rank(decay_linear(correlation(((self.high + self.low) / 2), adv40, 3).to_frame(), 6).CLOSE)
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    df.at[df['p1'] >= df['p2'], 'min'] = df['p2']
    df.at[df['p2'] >= df['p1'], 'min'] = df['p1']
    return df['min']
    # return min(rank(decay_linear(((((self.high + self.low) / 2) + self.high) - (self.vwap + self.high)).to_frame(), 20).CLOSE),rank(decay_linear(correlation(((self.high + self.low) / 2), adv40, 3).to_frame(), 6).CLOSE))


# Alpha#78	 (rank(correlation(sum(((low * 0.352233) + (vwap * (1 - 0.352233))), 19.7428),sum(adv40, 19.7428), 6.83313))^rank(correlation(rank(vwap), rank(volume), 5.77492)))
def alpha078(self):
    adv40 = sma(self.volume, 40)
    return (rank(
        correlation(ts_sum(((self.low * 0.352233) + (self.vwap * (1 - 0.352233))), 20), ts_sum(adv40, 20), 7)).pow(
        rank(correlation(rank(self.vwap), rank(self.volume), 6))))


# Alpha#79	 (rank(delta(IndNeutralize(((close * 0.60733) + (open * (1 - 0.60733))),IndClass.sector), 1.23438)) < rank(correlation(Ts_Rank(vwap, 3.60973), Ts_Rank(adv150,9.18637), 14.6644)))

# Alpha#80	 ((rank(Sign(delta(IndNeutralize(((open * 0.868128) + (high * (1 - 0.868128))),IndClass.industry), 4.04545)))^Ts_Rank(correlation(high, adv10, 5.11456), 5.53756)) * -1)

# Alpha#81	 ((rank(Log(product(rank((rank(correlation(vwap, sum(adv10, 49.6054),8.47743))^4)), 14.9655))) < rank(correlation(rank(vwap), rank(volume), 5.07914))) * -1)
def alpha081(self):
    adv10 = sma(self.volume, 10)
    return ((rank(log(product(rank((rank(correlation(self.vwap, ts_sum(adv10, 50), 8)).pow(4))), 15))) < rank(
        correlation(rank(self.vwap), rank(self.volume), 5))) * -1)


# Alpha#82	 (min(rank(decay_linear(delta(open, 1.46063), 14.8717)),Ts_Rank(decay_linear(correlation(IndNeutralize(volume, IndClass.sector), ((open * 0.634196) +(open * (1 - 0.634196))), 17.4842), 6.92131), 13.4283)) * -1)

# Alpha#83	 ((rank(delay(((high - low) / (sum(close, 5) / 5)), 2)) * rank(rank(volume))) / (((high -low) / (sum(close, 5) / 5)) / (vwap - close)))
def alpha083(self):
    return ((rank(delay(((self.high - self.low) / (ts_sum(self.close, 5) / 5)), 2)) * rank(rank(self.volume))) / (
            ((self.high - self.low) / (ts_sum(self.close, 5) / 5)) / (self.vwap - self.close)))


# Alpha#84	 SignedPower(Ts_Rank((vwap - ts_max(vwap, 15.3217)), 20.7127), delta(close,4.96796))
def alpha084(self):
    return pow(ts_rank((self.vwap - ts_max(self.vwap, 15)), 21), delta(self.close, 5))


# Alpha#85	 (rank(correlation(((high * 0.876703) + (close * (1 - 0.876703))), adv30,9.61331))^rank(correlation(Ts_Rank(((high + low) / 2), 3.70596), Ts_Rank(volume, 10.1595),7.11408)))
def alpha085(self):
    adv30 = sma(self.volume, 30)
    return (rank(correlation(((self.high * 0.876703) + (self.close * (1 - 0.876703))), adv30, 10)).pow(
        rank(correlation(ts_rank(((self.high + self.low) / 2), 4), ts_rank(self.volume, 10), 7))))


# Alpha#86	 ((Ts_Rank(correlation(close, sum(adv20, 14.7444), 6.00049), 20.4195) < rank(((open+ close) - (vwap + open)))) * -1)

def alpha086(self):
    adv20 = sma(self.volume, 20)
    return ((ts_rank(correlation(self.close, sma(adv20, 15), 6), 20) < rank(
        ((self.open + self.close) - (self.vwap + self.open)))) * -1)


# Alpha#87	 (max(rank(decay_linear(delta(((close * 0.369701) + (vwap * (1 - 0.369701))),1.91233), 2.65461)), Ts_Rank(decay_linear(abs(correlation(IndNeutralize(adv81,IndClass.industry), close, 13.4132)), 4.89768), 14.4535)) * -1)

# Alpha#88	 min(rank(decay_linear(((rank(open) + rank(low)) - (rank(high) + rank(close))),8.06882)), Ts_Rank(decay_linear(correlation(Ts_Rank(close, 8.44728), Ts_Rank(adv60,20.6966), 8.01266), 6.65053), 2.61957))
def alpha088(self):
    adv60 = sma(self.volume, 60)
    p1 = rank(decay_linear(((rank(self.open) + rank(self.low)) - (rank(self.high) + rank(self.close))).to_frame(),
                           8).CLOSE)
    p2 = ts_rank(decay_linear(correlation(ts_rank(self.close, 8), ts_rank(adv60, 21), 8).to_frame(), 7).CLOSE, 3)
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    df.at[df['p1'] >= df['p2'], 'min'] = df['p2']
    df.at[df['p2'] >= df['p1'], 'min'] = df['p1']
    return df['min']
    # return min(rank(decay_linear(((rank(self.open) + rank(self.low)) - (rank(self.high) + rank(self.close))).to_frame(),8).CLOSE), ts_rank(decay_linear(correlation(ts_rank(self.close, 8), ts_rank(adv60,20.6966), 8).to_frame(), 7).CLOSE, 3))


# Alpha#89	 (Ts_Rank(decay_linear(correlation(((low * 0.967285) + (low * (1 - 0.967285))), adv10,6.94279), 5.51607), 3.79744) - Ts_Rank(decay_linear(delta(IndNeutralize(vwap,IndClass.industry), 3.48158), 10.1466), 15.3012))

# Alpha#90	 ((rank((close - ts_max(close, 4.66719)))^Ts_Rank(correlation(IndNeutralize(adv40,IndClass.subindustry), low, 5.38375), 3.21856)) * -1)

# Alpha#91	 ((Ts_Rank(decay_linear(decay_linear(correlation(IndNeutralize(close,IndClass.industry), volume, 9.74928), 16.398), 3.83219), 4.8667) -rank(decay_linear(correlation(vwap, adv30, 4.01303), 2.6809))) * -1)

# Alpha#92	 min(Ts_Rank(decay_linear(((((high + low) / 2) + close) < (low + open)), 14.7221),18.8683), Ts_Rank(decay_linear(correlation(rank(low), rank(adv30), 7.58555), 6.94024),6.80584))
def alpha092(self):
    adv30 = sma(self.volume, 30)
    p1 = ts_rank(
        decay_linear(((((self.high + self.low) / 2) + self.close) < (self.low + self.open)).to_frame(), 15).CLOSE,
        19)
    p2 = ts_rank(decay_linear(correlation(rank(self.low), rank(adv30), 8).to_frame(), 7).CLOSE, 7)
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    df.at[df['p1'] >= df['p2'], 'min'] = df['p2']
    df.at[df['p2'] >= df['p1'], 'min'] = df['p1']
    return df['min']
    # return  min(ts_rank(decay_linear(((((self.high + self.low) / 2) + self.close) < (self.low + self.open)).to_frame(), 15).CLOSE,19), ts_rank(decay_linear(correlation(rank(self.low), rank(adv30), 8).to_frame(), 7).CLOSE,7))


# Alpha#93	 (Ts_Rank(decay_linear(correlation(IndNeutralize(vwap, IndClass.industry), adv81,17.4193), 19.848), 7.54455) / rank(decay_linear(delta(((close * 0.524434) + (vwap * (1 -0.524434))), 2.77377), 16.2664)))

# Alpha#94	 ((rank((vwap - ts_min(vwap, 11.5783)))^Ts_Rank(correlation(Ts_Rank(vwap,19.6462), Ts_Rank(adv60, 4.02992), 18.0926), 2.70756)) * -1)
def alpha094(self):
    adv60 = sma(self.volume, 60)
    return ((rank((self.vwap - ts_min(self.vwap, 12))).pow(
        ts_rank(correlation(ts_rank(self.vwap, 20), ts_rank(adv60, 4), 18), 3)) * -1))


# Alpha#95	 (rank((open - ts_min(open, 12.4105))) < Ts_Rank((rank(correlation(sum(((high + low)/ 2), 19.1351), sum(adv40, 19.1351), 12.8742))^5), 11.7584))
def alpha095(self):
    adv40 = sma(self.volume, 40)
    return (rank((self.open - ts_min(self.open, 12))) < ts_rank(
        (rank(correlation(sma(((self.high + self.low) / 2), 19), sma(adv40, 19), 13)).pow(5)), 12))


# Alpha#96	 (max(Ts_Rank(decay_linear(correlation(rank(vwap), rank(volume), 3.83878),4.16783), 8.38151), Ts_Rank(decay_linear(Ts_ArgMax(correlation(Ts_Rank(close, 7.45404),Ts_Rank(adv60, 4.13242), 3.65459), 12.6556), 14.0365), 13.4143)) * -1)
def alpha096(self):
    adv60 = sma(self.volume, 60)
    p1 = ts_rank(decay_linear(correlation(rank(self.vwap), rank(self.volume).to_frame(), 4), 4).CLOSE, 8)
    p2 = ts_rank(
        decay_linear(ts_argmax(correlation(ts_rank(self.close, 7), ts_rank(adv60, 4), 4), 13).to_frame(), 14).CLOSE,
        13)
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    df.at[df['p1'] >= df['p2'], 'max'] = df['p1']
    df.at[df['p2'] >= df['p1'], 'max'] = df['p2']
    return -1 * df['max']
    # return (max(ts_rank(decay_linear(correlation(rank(self.vwap), rank(self.volume).to_frame(), 4),4).CLOSE, 8), ts_rank(decay_linear(ts_argmax(correlation(ts_rank(self.close, 7),ts_rank(adv60, 4), 4), 13).to_frame(), 14).CLOSE, 13)) * -1)


# Alpha#97	 ((rank(decay_linear(delta(IndNeutralize(((low * 0.721001) + (vwap * (1 - 0.721001))),IndClass.industry), 3.3705), 20.4523)) - Ts_Rank(decay_linear(Ts_Rank(correlation(Ts_Rank(low,7.87871), Ts_Rank(adv60, 17.255), 4.97547), 18.5925), 15.7152), 6.71659)) * -1)

# Alpha#98	 (rank(decay_linear(correlation(vwap, sum(adv5, 26.4719), 4.58418), 7.18088)) -rank(decay_linear(Ts_Rank(Ts_ArgMin(correlation(rank(open), rank(adv15), 20.8187), 8.62571),6.95668), 8.07206)))
def alpha098(self):
    adv5 = sma(self.volume, 5)
    adv15 = sma(self.volume, 15)
    return (rank(decay_linear(correlation(self.vwap, sma(adv5, 26), 5).to_frame(), 7).CLOSE) - rank(
        decay_linear(ts_rank(ts_argmin(correlation(rank(self.open), rank(adv15), 21), 9), 7).to_frame(), 8).CLOSE))


class Alpha099(WQAlpha101):
    """
               [mul]
               /   \
           [lt]   -1
           /   \
       [rank] [rank]
       /        |
    [correlation] [correlation]
     /   |   \    /   |   \
    [sum] [sum] 8.8136 low volume 6.28259
     /   \    /   \
    [div] 19.8975 adv60 19.8975
     /   \
    high low

    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((rank(correlation(sum(((high + low) / 2), 19.8975), sum(adv60, 19.8975), 8.8136)) <rank(correlation(" \
               "low, volume, 6.28259))) * -1)"

    def forward(self):
        adv60 = ts_mean(self.volume, 60)
        self.data = (cs_rank(ts_corr(ts_sum(((self.high + self.low) / 2), 20), ts_sum(adv60, 20), 9)) <
                     cs_rank(ts_corr(self.low, self.volume, 6))) * -1
        return self.data


class Alpha101(WQAlpha101):
    """
           [div]
           /   \
       [sub]  [add]
       /   \   /   \
     close open [sub] .001
               /   \
             high  low
    """

    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return "((close - open) / ((high - low) + .001))"

    def forward(self):
        self.data = (self.close - self.open) / ((self.high - self.low) + 0.001)
        return self.data



from torchqtm._C._rolling import *
from pandas._libs.window.aggregations import (
    roll_rank,
    roll_quantile,
    roll_kurt,
    roll_var,
    roll_weighted_sum
)
from pandas._libs.algos import rank_1d, rank_2d
import numpy as np
from typing import Callable, Union
from numpy.typing import NDArray
import datetime
import functools


# PythonScalar = Union[str, float, bool]
# DatetimeLikeScalar = Union["Period", "Timestamp", "Timedelta"]
# PandasScalar = Union["Period", "Timestamp", "Timedelta", "Interval"]
# Scalar = Union[PythonScalar, PandasScalar, np.datetime64, np.timedelta64, datetime]


# @functools.lru_cache(maxsize=None)
def get_window_bounds(num_values, window_size):
    offset = 0
    end = np.arange(1 + offset, num_values + 1 + offset, 1, dtype=np.int64)
    start = end - window_size
    end = np.clip(end, 0, num_values)
    start = np.clip(start, 0, num_values)
    return start, end


def roll_apply(array: NDArray[np.float64],
               window_size: int,
               func: Callable[..., Union[np.float64, np.array]],
               *args,
               **kwargs) -> NDArray[np.float64]:
    start, end = get_window_bounds(len(array), window_size)
    if array.ndim == 1:
        return roll_apply_1D(array, start, end, window_size, func, args, kwargs)
    elif array.ndim == 2:
        return roll_apply_2D(array, start, end, window_size, func, args, kwargs)
    elif array.ndim == 3:
        return roll_apply_3D(array, start, end, window_size, func, args, kwargs)
    else:
        raise ValueError("Invalid number of dimensions")


def roll_apply_max(array: NDArray[np.float64],
                   window_size: int) -> NDArray[np.float64]:
    start, end = get_window_bounds(len(array), window_size)
    if array.ndim == 1:
        return roll_max(array, start, end, window_size)
    elif array.ndim == 2:
        if array.shape[1] <= 500 or window_size > 100:
            rlt = np.empty_like(array)
            for i in range(array.shape[1]):
                rlt[:, i] = roll_max(array[:, i], start, end, window_size)
            return rlt
        else:
            return roll_apply_2D(array, start, end, window_size, lambda x: np.nanmax(x, axis=0), (), {})
    else:
        raise ValueError("Invalid number of dimensions")


def roll_apply_min(array: NDArray[np.float64],
                   window_size: int) -> NDArray[np.float64]:
    start, end = get_window_bounds(len(array), window_size)
    if array.ndim == 1:
        return roll_min(array, start, end, window_size)
    elif array.ndim == 2:
        if array.shape[1] <= 500 or window_size > 100:  # 这里使用500是个经验法则了吧
            rlt = np.empty_like(array)
            for i in range(array.shape[1]):
                rlt[:, i] = roll_min(array[:, i], start, end, window_size)
            return rlt
        else:
            return roll_apply_2D(array, start, end, window_size, lambda x: np.nanmin(x, axis=0), (), {})
    else:
        raise ValueError("Invalid number of dimensions")


def roll_apply_mean(array: NDArray[np.float64],
                    window_size: int,
                    mode:str = "auto") -> NDArray[np.float64]:
    start, end = get_window_bounds(len(array), window_size)
    aux_func = roll_mean
    helper_func = lambda x: np.nanmean(x, axis=0)

    if array.ndim == 1:
        return aux_func(array, start, end, window_size)
    if mode == "auto":
        if array.ndim == 2:
            if window_size > 20 or array.shape[1] < 20:  # 这里使用500是个经验法则了吧
                rlt = np.empty_like(array)
                for i in range(array.shape[1]):
                    rlt[:, i] = aux_func(array[:, i], start, end, window_size)
                return rlt
            else:
                return roll_apply_2D(array, start, end, window_size, helper_func, (), {})
        else:
            raise ValueError("Invalid number of dimensions")

    elif mode == "overall":
        return roll_apply_2D(array, start, end, window_size, helper_func, (), {})

    elif mode == "single":
        if array.ndim == 2:
            rlt = np.empty_like(array)
            for i in range(array.shape[1]):
                rlt[:, i] = aux_func(array[:, i], start, end, window_size)
            return rlt


def roll_apply_sum(array: NDArray[np.float64],
                   window_size: int,
                   mode: str = "auto") -> NDArray[np.float64]:
    start, end = get_window_bounds(len(array), window_size)
    aux_func = roll_sum
    helper_func = lambda x: np.nansum(x, axis=0)

    if array.ndim == 1:
        return aux_func(array)
    if mode == "auto":
        if array.ndim == 2:
            if window_size > 50:  # 这里使用500是个经验法则了吧
                rlt = np.empty_like(array)
                for i in range(array.shape[1]):
                    rlt[:, i] = aux_func(array[:, i], start, end, window_size)
                return rlt
            else:
                return roll_apply_2D(array, start, end, window_size, helper_func, (), {})
        else:
            raise ValueError("Invalid number of dimensions")

    elif mode == "overall":
        return roll_apply_2D(array, start, end, window_size, helper_func, (), {})

    elif mode == "single":
        if array.ndim == 2:
            rlt = np.empty_like(array)
            for i in range(array.shape[1]):
                rlt[:, i] = aux_func(array[:, i], start, end, window_size)
            return rlt


def roll_apply_rank(array: NDArray[np.float64],
                    window_size: int,
                    method="min",
                    percentile: bool = False,
                    ascending: bool = True,
                    mode: str = "auto") -> NDArray[np.float64]:
    start, end = get_window_bounds(len(array), window_size)
    aux_func = roll_rank
    helper_func = lambda x: rank_2d(x[::-1], axis=0)[-1]
    if array.ndim == 1:
        return aux_func(array, start, end, window_size, percentile, method, ascending)
    if mode == "auto":
        if array.ndim == 2:
            if window_size > 20:
                rlt = np.empty_like(array)
                for i in range(array.shape[1]):
                    rlt[:, i] = aux_func(array[:, i], start, end, window_size, percentile, method, ascending)
                return rlt
            else:
                return roll_apply_2D(array, start, end, window_size, helper_func, (), {})
        else:
            raise ValueError("Invalid number of dimensions")

    elif mode == "overall":
        return roll_apply_2D(array, start, end, window_size, helper_func, (), {})

    elif mode == "single":
        if array.ndim == 2:
            rlt = np.empty_like(array)
            for i in range(array.shape[1]):
                rlt[:, i] = aux_func(array[:, i], start, end, window_size, percentile, method, ascending)
            return rlt





"""The underlying rawdata structure used in gplearn.

The :mod:`gplearn._program` module contains the underlying representation of a
computer program. It is used for creating and evolving programs used in the
:mod:`gplearn.genetic` module.
"""

# Author: Trevor Stephens <trevorstephens.com>
#
# License: BSD 3 clause

from copy import copy

import numpy as np
from sklearn.utils.random import sample_without_replacement

from .functions import _Function
from .utils import check_random_state


class _Program(object):

    """A program-like representation of the evolved program.

    This is the underlying rawdata-structure used by the public classes in the
    :mod:`gplearn.genetic` module. It should not be used directly by the user.

    Parameters
    ----------
    function_set : list
        A list of valid functions to use in the program.

    arities : dict
        A dictionary of the form `{arity: [functions]}`. The arity is the
        number of arguments that the function takes, the functions must match
        those in the `function_set` parameter.

    init_depth : tuple of two ints
        The range of tree depths for the initial population of naive formulas.
        Individual trees will randomly choose a maximum depth from this range.
        When combined with `init_method='half and half'` this yields the well-
        known 'ramped half and half' initialization method.

    init_method : str
        - 'grow' : Nodes are chosen at random from both functions and
          terminals, allowing for smaller trees than `init_depth` allows. Tends
          to grow asymmetrical trees.
        - 'full' : Functions are chosen until the `init_depth` is reached, and
          then terminals are selected. Tends to grow 'bushy' trees.
        - 'half and half' : Trees are grown through a 50/50 mix of 'full' and
          'grow', making for a mix of tree shapes in the initial population.

    n_features : int
        The number of features in `X`.

    const_range : tuple of two floats
        The range of constants to include in the formulas.

    metric : _Fitness object
        The raw fitness metric.

    p_point_replace : float
        The probability that any given node will be mutated during point
        mutation.

    parsimony_coefficient : float
        This constant penalizes large programs by adjusting their fitness to
        be less favorable for selection. Larger values penalize the program
        more which can control the phenomenon known as 'bloat'. Bloat is when
        evolution is increasing the size of programs without a significant
        increase in fitness, which is costly for computation time and makes for
        a less understandable final result. This parameter may need to be tuned
        over successive runs.

    random_state : RandomState instance
        The random number generator. Note that ints, or None are not allowed.
        The reason for this being passed is that during parallel evolution the
        same program object may be accessed by multiple parallel processes.

    transformer : _Function object, optional (default=None)
        The function to transform the output of the program to probabilities,
        only used for the SymbolicClassifier.

    feature_names : list, optional (default=None)
        Optional list of feature names, used purely for representations in
        the `print` operation or `export_graphviz`. If None, then X0, X1, etc
        will be used for representations.

    program : list, optional (default=None)
        The flattened tree representation of the program. If None, a new naive
        random tree will be grown. If provided, it will be validated.

    Attributes
    ----------
    program : list
        The flattened tree representation of the program.

    raw_fitness_ : float
        The raw fitness of the individual program.

    fitness_ : float
        The penalized fitness of the individual program.

    oob_fitness_ : float
        The out-of-bag raw fitness of the individual program for the held-out
        samples. Only present when sub-sampling was used in the estimator by
        specifying `max_samples` < 1.0.

    parents : dict, or None
        If None, this is a naive random program from the initial population.
        Otherwise it includes meta-rawdata about the program's parent(s) as well
        as the genetic operations performed to yield the current program. This
        is set outside this class by the controlling evolution loops.

    depth_ : int
        The maximum depth of the program tree.

    length_ : int
        The number of functions and terminals in the program.

    """

    def __init__(self,
                 function_set,
                 arities,
                 init_depth,
                 init_method,
                 n_features,
                 const_range,
                 metric,
                 p_point_replace,
                 parsimony_coefficient,
                 random_state,
                 transformer=None,
                 feature_names=None,
                 program=None,
                 strategy=None):

        self.function_set = function_set
        self.arities = arities
        self.init_depth = (init_depth[0], init_depth[1] + 1)
        self.init_method = init_method
        self.n_features = n_features
        self.const_range = const_range
        self.metric = metric
        self.p_point_replace = p_point_replace
        self.parsimony_coefficient = parsimony_coefficient
        self.transformer = transformer
        self.feature_names = feature_names
        self.program = program
        self.strategy = strategy
        if self.program is not None:
            if not self.validate_program():
                raise ValueError('The supplied program is incomplete.')
        else:
            # Create a naive random program
            self.program = self.build_program(random_state)

        self.raw_fitness_ = None
        self.fitness_ = None
        self.parents = None
        self._n_samples = None
        self._max_samples = None
        self._indices_state = None

    def build_program(self, random_state):
        """Build a naive random program.

        Parameters
        ----------
        random_state : RandomState instance
            The random number generator.

        Returns
        -------
        program : list
            The flattened tree representation of the program.

        """
        if self.init_method == 'half and half':
            method = ('full' if random_state.randint(2) else 'grow')
        else:
            method = self.init_method
        max_depth = random_state.randint(*self.init_depth)

        # Start a program with a function to avoid degenerative programs
        function = random_state.randint(len(self.function_set))
        function = self.function_set[function]
        program = [function]
        terminal_stack = [function.arity]

        while terminal_stack:
            depth = len(terminal_stack)
            choice = self.n_features + len(self.function_set)
            choice = random_state.randint(choice)
            # Determine if we are adding a function or terminal
            if (depth < max_depth) and (method == 'full' or
                                        choice <= len(self.function_set)):
                function = random_state.randint(len(self.function_set))
                function = self.function_set[function]
                program.append(function)
                terminal_stack.append(function.arity)
            else:
                # We need a terminal, add a variable or constant
                if self.const_range is not None:
                    terminal = random_state.randint(self.n_features + 1)
                else:
                    terminal = random_state.randint(self.n_features)
                if terminal == self.n_features:
                    terminal = random_state.uniform(*self.const_range)
                    if self.const_range is None:
                        # We should never get here
                        raise ValueError('A constant was produced with '
                                         'const_range=None.')
                program.append(terminal)
                terminal_stack[-1] -= 1
                while terminal_stack[-1] == 0:
                    terminal_stack.pop()
                    if not terminal_stack:
                        return program
                    terminal_stack[-1] -= 1

        # We should never get here
        return None

    def validate_program(self):
        """Rough check that the embedded program in the object is valid."""
        terminals = [0]
        for node in self.program:
            if isinstance(node, _Function):
                terminals.append(node.arity)
            else:
                terminals[-1] -= 1
                while terminals[-1] == 0:
                    terminals.pop()
                    terminals[-1] -= 1
        return terminals == [-1]

    def __str__(self):
        """Overloads `print` output of the object to resemble a LISP tree."""
        terminals = [0]
        output = ''
        for i, node in enumerate(self.program):
            if isinstance(node, _Function):
                terminals.append(node.arity)
                output += node.name + '('
            else:
                if isinstance(node, int):
                    if self.feature_names is None:
                        output += 'X%s' % node
                    else:
                        output += self.feature_names[node]
                else:
                    output += '%.3f' % node
                terminals[-1] -= 1
                while terminals[-1] == 0:
                    terminals.pop()
                    terminals[-1] -= 1
                    output += ')'
                if i != len(self.program) - 1:
                    output += ', '
        return output

    def export_graphviz(self, fade_nodes=None):
        """Returns a string, Graphviz script for visualizing the program.

        Parameters
        ----------
        fade_nodes : list, optional
            A list of node indices to fade out for showing which were removed
            during evolution.

        Returns
        -------
        output : string
            The Graphviz script to plot the tree representation of the program.

        """
        terminals = []
        if fade_nodes is None:
            fade_nodes = []
        output = 'digraph program {\nnode [style=filled]\n'
        for i, node in enumerate(self.program):
            fill = '#cecece'
            if isinstance(node, _Function):
                if i not in fade_nodes:
                    fill = '#136ed4'
                terminals.append([node.arity, i])
                output += ('%d [label="%s", fillcolor="%s"] ;\n'
                           % (i, node.name, fill))
            else:
                if i not in fade_nodes:
                    fill = '#60a6f6'
                if isinstance(node, int):
                    if self.feature_names is None:
                        feature_name = 'X%s' % node
                    else:
                        feature_name = self.feature_names[node]
                    output += ('%d [label="%s", fillcolor="%s"] ;\n'
                               % (i, feature_name, fill))
                else:
                    output += ('%d [label="%.3f", fillcolor="%s"] ;\n'
                               % (i, node, fill))
                if i == 0:
                    # A degenerative program of only one node
                    return output + '}'
                terminals[-1][0] -= 1
                terminals[-1].append(i)
                while terminals[-1][0] == 0:
                    output += '%d -> %d ;\n' % (terminals[-1][1],
                                                terminals[-1][-1])
                    terminals[-1].pop()
                    if len(terminals[-1]) == 2:
                        parent = terminals[-1][-1]
                        terminals.pop()
                        if not terminals:
                            return output + '}'
                        terminals[-1].append(parent)
                        terminals[-1][0] -= 1

        # We should never get here
        return None

    def _depth(self):
        """Calculates the maximum depth of the program tree."""
        terminals = [0]
        depth = 1
        for node in self.program:
            if isinstance(node, _Function):
                terminals.append(node.arity)
                depth = max(len(terminals), depth)
            else:
                terminals[-1] -= 1
                while terminals[-1] == 0:
                    terminals.pop()
                    terminals[-1] -= 1
        return depth - 1

    def _length(self):
        """Calculates the number of functions and terminals in the program."""
        return len(self.program)

    def execute(self, X):
        """Execute the program according to X.

        Parameters
        ----------
        X : {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        y_hats : array-like, shape = [n_samples]
            The result of executing the program on X.

        """
        # Check for single-node programs
        node = self.program[0]
        if isinstance(node, float):
        # TODO: 非常需要改
            return np.repeat(node, X.shape[-2])
        if isinstance(node, int):
            return X[..., node]

        apply_stack = []

        for node in self.program:

            if isinstance(node, _Function):
                apply_stack.append([node])
            else:
                # Lazily evaluate later
                apply_stack[-1].append(node)

            while len(apply_stack[-1]) == apply_stack[-1][0].arity + 1:
                # Apply functions that have sufficient arguments
                function = apply_stack[-1][0]
                terminals = [np.reshape(np.repeat(t, np.prod(X.shape[:-1])), X.shape[:-1]) if isinstance(t,float)
                             else X[..., t] if isinstance(t, int)
                             else t for t in apply_stack[-1][1:]]
                intermediate_result = function(*terminals)
                if len(apply_stack) != 1:
                    apply_stack.pop()
                    apply_stack[-1].append(intermediate_result)
                else:
                    return intermediate_result

        # We should never get here
        return None

    def get_all_indices(self, n_samples=None, max_samples=None,
                        random_state=None):
        """Get the indices on which to evaluate the fitness of a program.

        Parameters
        ----------
        n_samples : int
            The number of samples.

        max_samples : int
            The maximum number of samples to use.

        random_state : RandomState instance
            The random number generator.

        Returns
        -------
        indices : array-like, shape = [n_samples]
            The in-sample indices.

        not_indices : array-like, shape = [n_samples]
            The out-of-sample indices.

        """
        if self._indices_state is None and random_state is None:
            raise ValueError('The program has not been evaluated for fitness '
                             'yet, indices not available.')

        if n_samples is not None and self._n_samples is None:
            self._n_samples = n_samples
        if max_samples is not None and self._max_samples is None:
            self._max_samples = max_samples
        if random_state is not None and self._indices_state is None:
            self._indices_state = random_state.get_state()

        indices_state = check_random_state(None)
        indices_state.set_state(self._indices_state)

        not_indices = sample_without_replacement(
            self._n_samples,
            self._n_samples - self._max_samples,
            random_state=indices_state)
        sample_counts = np.bincount(not_indices, minlength=self._n_samples)
        indices = np.where(sample_counts == 0)[0]

        return indices, not_indices

    def _indices(self):
        """Get the indices used to measure the program's fitness."""
        return self.get_all_indices()[0]

    def raw_fitness(self, X, y, sample_weight):
        """Evaluate the raw fitness of the program according to X, y.

        Parameters
        ----------
        X : {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target values.

        sample_weight : array-like, shape = [n_samples]
            Weights applied to individual samples.

        Returns
        -------
        raw_fitness : float
            The raw fitness of the program.

        """
        y_pred = self.execute(X)
        if self.transformer:
            y_pred = self.transformer(y_pred)

        try:
            raw_fitness = self.metric(y, y_pred, sample_weight)
        except TypeError:  # TypeError 说明需要传入strategy参数
            raw_fitness = self.metric(y, y_pred, sample_weight, self.strategy)

        return raw_fitness

    def fitness(self, parsimony_coefficient=None):
        """Evaluate the penalized fitness of the program according to X, y.

        Parameters
        ----------
        parsimony_coefficient : float, optional
            If automatic parsimony is being used, the computed value according
            to the population. Otherwise the initialized value is used.

        Returns
        -------
        fitness : float
            The penalized fitness of the program.

        """
        if parsimony_coefficient is None:
            parsimony_coefficient = self.parsimony_coefficient
        penalty = parsimony_coefficient * len(self.program) * self.metric.sign
        return self.raw_fitness_ - penalty

    def get_subtree(self, random_state, program=None):
        """Get a random subtree from the program.

        Parameters
        ----------
        random_state : RandomState instance
            The random number generator.

        program : list, optional (default=None)
            The flattened tree representation of the program. If None, the
            embedded tree in the object will be used.

        Returns
        -------
        start, end : tuple of two ints
            The indices of the start and end of the random subtree.

        """
        if program is None:
            program = self.program
        # Choice of crossover points follows Koza's (1992) widely used approach
        # of choosing functions 90% of the time and leaves 10% of the time.
        probs = np.array([0.9 if isinstance(node, _Function) else 0.1
                          for node in program])
        probs = np.cumsum(probs / probs.sum())
        start = np.searchsorted(probs, random_state.uniform())

        stack = 1
        end = start
        while stack > end - start:
            node = program[end]
            if isinstance(node, _Function):
                stack += node.arity
            end += 1

        return start, end

    def reproduce(self):
        """Return a copy of the embedded program."""
        return copy(self.program)

    def crossover(self, donor, random_state):
        """Perform the crossover genetic operation on the program.

        Crossover selects a random subtree from the embedded program to be
        replaced. A donor also has a subtree selected at random and this is
        inserted into the original parent to form an offspring.

        Parameters
        ----------
        donor : list
            The flattened tree representation of the donor program.

        random_state : RandomState instance
            The random number generator.

        Returns
        -------
        program : list
            The flattened tree representation of the program.

        """
        # Get a subtree to replace
        start, end = self.get_subtree(random_state)
        removed = range(start, end)
        # Get a subtree to donate
        donor_start, donor_end = self.get_subtree(random_state, donor)
        donor_removed = list(set(range(len(donor))) -
                             set(range(donor_start, donor_end)))
        # Insert genetic material from donor
        return (self.program[:start] +
                donor[donor_start:donor_end] +
                self.program[end:]), removed, donor_removed

    def subtree_mutation(self, random_state):
        """Perform the subtree mutation operation on the program.

        Subtree mutation selects a random subtree from the embedded program to
        be replaced. A donor subtree is generated at random and this is
        inserted into the original parent to form an offspring. This
        implementation uses the "headless chicken" method where the donor
        subtree is grown using the initialization methods and a subtree of it
        is selected to be donated to the parent.

        Parameters
        ----------
        random_state : RandomState instance
            The random number generator.

        Returns
        -------
        program : list
            The flattened tree representation of the program.

        """
        # Build a new naive program
        chicken = self.build_program(random_state)
        # Do subtree mutation via the headless chicken method!
        return self.crossover(chicken, random_state)

    def hoist_mutation(self, random_state):
        """Perform the hoist mutation operation on the program.

        Hoist mutation selects a random subtree from the embedded program to
        be replaced. A random subtree of that subtree is then selected and this
        is 'hoisted' into the original subtrees location to form an offspring.
        This method helps to control bloat.

        Parameters
        ----------
        random_state : RandomState instance
            The random number generator.

        Returns
        -------
        program : list
            The flattened tree representation of the program.

        """
        # Get a subtree to replace
        start, end = self.get_subtree(random_state)
        subtree = self.program[start:end]
        # Get a subtree of the subtree to hoist
        sub_start, sub_end = self.get_subtree(random_state, subtree)
        hoist = subtree[sub_start:sub_end]
        # Determine which nodes were removed for plotting
        removed = list(set(range(start, end)) -
                       set(range(start + sub_start, start + sub_end)))
        return self.program[:start] + hoist + self.program[end:], removed

    def point_mutation(self, random_state):
        """Perform the point mutation operation on the program.

        Point mutation selects random nodes from the embedded program to be
        replaced. Terminals are replaced by other terminals and functions are
        replaced by other functions that require the same number of arguments
        as the original node. The resulting tree forms an offspring.

        Parameters
        ----------
        random_state : RandomState instance
            The random number generator.

        Returns
        -------
        program : list
            The flattened tree representation of the program.

        """
        program = copy(self.program)

        # Get the nodes to modify
        mutate = np.where(random_state.uniform(size=len(program)) <
                          self.p_point_replace)[0]

        for node in mutate:
            if isinstance(program[node], _Function):
                arity = program[node].arity
                # Find a valid replacement with same arity
                replacement = len(self.arities[arity])
                replacement = random_state.randint(replacement)
                replacement = self.arities[arity][replacement]
                program[node] = replacement
            else:
                # We've got a terminal, add a const or variable
                if self.const_range is not None:
                    terminal = random_state.randint(self.n_features + 1)
                else:
                    terminal = random_state.randint(self.n_features)
                if terminal == self.n_features:
                    terminal = random_state.uniform(*self.const_range)
                    if self.const_range is None:
                        # We should never get here
                        raise ValueError('A constant was produced with '
                                         'const_range=None.')
                program[node] = terminal

        return program, list(mutate)

    depth_ = property(_depth)
    length_ = property(_length)
    indices_ = property(_indices)



import pandas as pd
import numpy as np
from bisect import bisect_right

import typing


def convert_date_to_int(dt):
    t = dt.year * 10000 + dt.month * 100 + dt.day
    t *= 1000000
    return t


def convert_dates_to_int(dates: pd.DatetimeIndex):
    intDates = dates.year * 10000 + dates.month * 100 + dates.day
    intDates *= 10e6
    return intDates.values.astype(np.uint64)


PRICE_FIELDS = {
    'open', 'close', 'high', 'low', 'limit_up', 'limit_down', 'acc_net_value', 'unit_net_value'
}

FIELDS_REQUIRE_ADJUSTMENT = set(list(PRICE_FIELDS) + ['volume'])


def _factor_for_date(dates, factors, d):
    pos = bisect_right(dates, d)
    return factors[pos-1]

# array([(             0,   1.     ), (19910502000000,   1.40968),
#        (19910817000000,   1.63169), (19920323000000,   2.46272),
#        (19930524000000,   4.6701 ), (19940711000000,   7.47216),
#        (19950925000000,   9.24767), (19960527000000,  18.4953 ),
#        (19970825000000,  27.9199 ), (19991018000000,  28.6789 ),
#        (20001106000000,  22.9762 ), (20020723000000,  23.2163 ),
#        (20030929000000,  23.5953 ), (20070620000000,  25.9629 ),
#        (20081031000000,  33.8521 ), (20121019000000,  34.1045 ),
#        (20130620000000,  55.0537 ), (20140612000000,  66.9741 ),
#        (20150413000000,  81.0815 ), (20160616000000,  98.7449 ),
#        (20170721000000, 100.188  ), (20180712000000, 101.764  ),
#        (20190626000000, 102.875  ), (20200528000000, 104.629  ),
#        (20210514000000, 105.452  ), (20220722000000, 107.333  )],
#       dtype=[('start_date', '<i8'), ('ex_cum_factor', '<f8')])


def adjust_bars(
        bars: np.ndarray,
        ex_factors: np.ndarray,
        fields: typing.Iterable[str],
        adjust_type: str,
        adjust_origin: pd.Timestamp,
):
    """
    Parameters
    ----------
    bars : np.ndarray
    ex_factors : np.ndarray
    fields : 不太确定
    adjust_type : str
    adjust_origin: pd.Timestamp
    """

    if ex_factors is None or len(bars) == 0:
        return bars

    dates = ex_factors['start_date']
    ex_cum_factors = ex_factors['ex_cum_factor']

    if adjust_type == 'pre':
        adjust_orig_dt = np.int64(convert_date_to_int(adjust_origin))
        base_adjust_rate = _factor_for_date(dates, ex_cum_factors, adjust_orig_dt)
    else:
        base_adjust_rate = 1.0

    start_date = bars['datetime'][0]
    end_date = bars['datetime'][-1]

    if (_factor_for_date(dates, ex_cum_factors, start_date) == base_adjust_rate and
            _factor_for_date(dates, ex_cum_factors, end_date) == base_adjust_rate):
        return bars

    factors = ex_cum_factors.take(dates.searchsorted(bars['datetime'], side='right') - 1)

    # 复权
    bars = np.copy(bars)
    factors /= base_adjust_rate
    if isinstance(fields, str):
        if fields in PRICE_FIELDS:
            bars[fields] *= factors
            return bars
        elif fields == 'volume':
            bars[fields] *= (1 / factors)
            return bars
        # should not got here
        return bars

    for f in bars.dtype.names:
        if f in PRICE_FIELDS:
            bars[f] *= factors
        elif f == 'volume':
            bars[f] *= (1 / factors)
    return bars



import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.op.functional import *
import pandas as pd
import numpy as np
from abc import ABCMeta, abstractmethod
from torchqtm.tdbt.backtest import BackTestEnv
from torchqtm.base import BaseAlpha


class MLAlpha(BaseAlpha):
    def __init__(self, env: BackTestEnv, model=None, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'machine learning alpha'
        self.model = model

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError




